from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode

from nodetool.nodes.replicate.image.generate import Pixel
from nodetool.nodes.replicate.image.generate import Product_size

class AdInpaint(GraphNode):
    pixel: Pixel | GraphNode | tuple[GraphNode, str] = Field(default='512 * 512', description='image total pixel')
    scale: int | GraphNode | tuple[GraphNode, str] = Field(default=3, description='Factor to scale image by (maximum: 4)')
    prompt: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Product name or prompt')
    image_num: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description='Number of image to generate')
    image_path: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='input image')
    manual_seed: int | GraphNode | tuple[GraphNode, str] = Field(default=-1, description='Manual Seed')
    product_size: Product_size | GraphNode | tuple[GraphNode, str] = Field(default='Original', description='Max product size')
    guidance_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=7.5, description='Guidance Scale')
    negative_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='low quality, out of frame, illustration, 3d, sepia, painting, cartoons, sketch, watermark, text, Logo, advertisement', description="Anything you don't want in the photo")
    num_inference_steps: int | GraphNode | tuple[GraphNode, str] = Field(default=20, description='Inference Steps')
    @classmethod
    def get_node_type(cls): return "replicate.image.generate.AdInpaint"



class ControlnetRealisticVision(GraphNode):
    seed: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Leave blank to randomize')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image')
    steps: int | GraphNode | tuple[GraphNode, str] = Field(default=20, description=' num_inference_steps')
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='(a tabby cat)+++, high resolution, sitting on a park bench', description=None)
    strength: float | GraphNode | tuple[GraphNode, str] = Field(default=0.8, description='control strength/weight')
    max_width: float | GraphNode | tuple[GraphNode, str] = Field(default=612, description='max width of mask/image')
    max_height: float | GraphNode | tuple[GraphNode, str] = Field(default=612, description='max height of mask/image')
    guidance_scale: int | GraphNode | tuple[GraphNode, str] = Field(default=10, description='guidance_scale')
    negative_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='(deformed iris, deformed pupils, semi-realistic, cgi, 3d, render, sketch, cartoon, drawing, anime:1.4), text, close up, cropped, out of frame, worst quality, low quality, jpeg artifacts, ugly, duplicate, morbid, mutilated, extra fingers, mutated hands, poorly drawn hands, poorly drawn face, mutation, deformed, blurry, dehydrated, bad anatomy, bad proportions, extra limbs, cloned face, disfigured, gross proportions, malformed limbs, missing arms, missing legs, extra arms, extra legs, fused fingers, too many fingers, long neck', description=None)
    @classmethod
    def get_node_type(cls): return "replicate.image.generate.ControlnetRealisticVision"


from nodetool.nodes.replicate.image.generate import Sizing_strategy

class Illusions(GraphNode):
    seed: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description=None)
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Optional img2img')
    width: int | GraphNode | tuple[GraphNode, str] = Field(default=768, description=None)
    height: int | GraphNode | tuple[GraphNode, str] = Field(default=768, description=None)
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='a painting of a 19th century town', description=None)
    mask_image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Optional mask for inpainting')
    num_outputs: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description='Number of outputs')
    control_image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Control image')
    controlnet_end: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='When controlnet conditioning ends')
    guidance_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=7.5, description='Scale for classifier-free guidance')
    negative_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='ugly, disfigured, low quality, blurry, nsfw', description='The negative prompt to guide image generation.')
    prompt_strength: float | GraphNode | tuple[GraphNode, str] = Field(default=0.8, description='Prompt strength when using img2img / inpaint. 1.0 corresponds to full destruction of information in image')
    sizing_strategy: Sizing_strategy | GraphNode | tuple[GraphNode, str] = Field(default='width/height', description='Decide how to resize images – use width/height, resize based on input image or control image')
    controlnet_start: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description='When controlnet conditioning starts')
    num_inference_steps: int | GraphNode | tuple[GraphNode, str] = Field(default=40, description='Number of diffusion steps')
    controlnet_conditioning_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=0.75, description='How strong the controlnet conditioning is')
    @classmethod
    def get_node_type(cls): return "replicate.image.generate.Illusions"


from nodetool.nodes.replicate.image.generate import Scheduler

class Juggernaut_XL_V9(GraphNode):
    seed: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Random seed. Leave blank to randomize the seed')
    width: int | GraphNode | tuple[GraphNode, str] = Field(default=1024, description='Width of output image')
    height: int | GraphNode | tuple[GraphNode, str] = Field(default=1024, description='Height of output image')
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='beautiful lady, (freckles), big smile, ruby eyes, short hair, dark makeup, hyperdetailed photography, soft light, head and shoulders portrait, cover', description='Input prompt')
    scheduler: Scheduler | GraphNode | tuple[GraphNode, str] = Field(default='DPM++SDE', description='scheduler')
    num_outputs: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description='Number of images to output.')
    guidance_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=2, description='Scale for classifier-free guidance')
    apply_watermark: bool | GraphNode | tuple[GraphNode, str] = Field(default=True, description='Applies a watermark to enable determining if an image is generated in downstream applications. If you have other provisions for generating or deploying images safely, you can use this to disable watermarking.')
    negative_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='CGI, Unreal, Airbrushed, Digital', description='Input Negative Prompt')
    num_inference_steps: int | GraphNode | tuple[GraphNode, str] = Field(default=5, description='Number of denoising steps')
    disable_safety_checker: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='Disable safety checker for generated images. This feature is only available through the API. See [https://replicate.com/docs/how-does-replicate-work#safety](https://replicate.com/docs/how-does-replicate-work#safety)')
    @classmethod
    def get_node_type(cls): return "replicate.image.generate.Juggernaut_XL_V9"


from nodetool.nodes.replicate.image.generate import Width
from nodetool.nodes.replicate.image.generate import Height
from nodetool.nodes.replicate.image.generate import Output_format

class Kandinsky(GraphNode):
    seed: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Random seed. Leave blank to randomize the seed')
    width: Width | GraphNode | tuple[GraphNode, str] = Field(default=512, description='Width of output image. Lower the setting if hits memory limits.')
    height: Height | GraphNode | tuple[GraphNode, str] = Field(default=512, description='Height of output image. Lower the setting if hits memory limits.')
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='A moss covered astronaut with a black background', description='Input prompt')
    num_outputs: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description='Number of images to output.')
    output_format: Output_format | GraphNode | tuple[GraphNode, str] = Field(default='webp', description='Output image format')
    negative_prompt: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Specify things to not see in the output')
    num_inference_steps: int | GraphNode | tuple[GraphNode, str] = Field(default=75, description='Number of denoising steps')
    num_inference_steps_prior: int | GraphNode | tuple[GraphNode, str] = Field(default=25, description='Number of denoising steps for priors')
    @classmethod
    def get_node_type(cls): return "replicate.image.generate.Kandinsky"


from nodetool.nodes.replicate.image.generate import Scheduler

class PlaygroundV2(GraphNode):
    mask: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Input mask for inpaint mode. Black areas will be preserved, white areas will be inpainted.')
    seed: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Random seed. Leave blank to randomize the seed')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image for img2img or inpaint mode')
    width: int | GraphNode | tuple[GraphNode, str] = Field(default=1024, description='Width of output image')
    height: int | GraphNode | tuple[GraphNode, str] = Field(default=1024, description='Height of output image')
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='Astronaut in a jungle, cold color palette, muted colors, detailed, 8k', description='Input prompt')
    scheduler: Scheduler | GraphNode | tuple[GraphNode, str] = Field(default='DPMSolver++', description='Scheduler. DPMSolver++ or DPM++2MKarras is recommended for most cases')
    num_outputs: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description='Number of images to output.')
    guidance_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=3, description='Scale for classifier-free guidance')
    apply_watermark: bool | GraphNode | tuple[GraphNode, str] = Field(default=True, description='Applies a watermark to enable determining if an image is generated in downstream applications. If you have other provisions for generating or deploying images safely, you can use this to disable watermarking.')
    negative_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='ugly, deformed, noisy, blurry, distorted', description='Negative Input prompt')
    prompt_strength: float | GraphNode | tuple[GraphNode, str] = Field(default=0.8, description='Prompt strength when using img2img / inpaint. 1.0 corresponds to full destruction of information in image')
    num_inference_steps: int | GraphNode | tuple[GraphNode, str] = Field(default=25, description='Number of denoising steps')
    disable_safety_checker: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='Disable safety checker for generated images. This feature is only available through the API. See https://replicate.com/docs/how-does-replicate-work#safety')
    @classmethod
    def get_node_type(cls): return "replicate.image.generate.PlaygroundV2"


from nodetool.nodes.replicate.image.generate import Scheduler

class Proteus(GraphNode):
    mask: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input mask for inpaint mode. Black areas will be preserved, white areas will be inpainted.')
    seed: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Random seed. Leave blank to randomize the seed')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image for img2img or inpaint mode')
    width: int | GraphNode | tuple[GraphNode, str] = Field(default=1024, description='Width of output image. Recommended 1024 or 1280')
    height: int | GraphNode | tuple[GraphNode, str] = Field(default=1024, description='Height of output image. Recommended 1024 or 1280')
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='black fluffy gorgeous dangerous cat animal creature, large orange eyes, big fluffy ears, piercing gaze, full moon, dark ambiance, best quality, extremely detailed', description='Input prompt')
    scheduler: Scheduler | GraphNode | tuple[GraphNode, str] = Field(default='DPM++2MSDE', description='scheduler')
    num_outputs: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description='Number of images to output.')
    guidance_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=7.5, description='Scale for classifier-free guidance. Recommended 4-6')
    apply_watermark: bool | GraphNode | tuple[GraphNode, str] = Field(default=True, description='Applies a watermark to enable determining if an image is generated in downstream applications. If you have other provisions for generating or deploying images safely, you can use this to disable watermarking.')
    negative_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='nsfw, bad quality, bad anatomy, worst quality, low quality, low resolutions, extra fingers, blur, blurry, ugly, wrongs proportions, watermark, image artifacts, lowres, ugly, jpeg artifacts, deformed, noisy image', description='Negative Input prompt')
    prompt_strength: float | GraphNode | tuple[GraphNode, str] = Field(default=0.8, description='Prompt strength when using img2img / inpaint. 1.0 corresponds to full destruction of information in image')
    num_inference_steps: int | GraphNode | tuple[GraphNode, str] = Field(default=20, description='Number of denoising steps. 20 to 60 steps for more detail, 20 steps for faster results.')
    disable_safety_checker: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='Disable safety checker for generated images. This feature is only available through the API. See https://replicate.com/docs/how-does-replicate-work#safety')
    @classmethod
    def get_node_type(cls): return "replicate.image.generate.Proteus"


from nodetool.nodes.replicate.image.generate import Refine
from nodetool.nodes.replicate.image.generate import Scheduler
from nodetool.nodes.replicate.image.generate import Controlnet_1
from nodetool.nodes.replicate.image.generate import Controlnet_2
from nodetool.nodes.replicate.image.generate import Controlnet_3
from nodetool.nodes.replicate.image.generate import Sizing_strategy

class RealVisXLV3(GraphNode):
    mask: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input mask for inpaint mode. Black areas will be preserved, white areas will be inpainted.')
    seed: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Random seed. Leave blank to randomize the seed')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image for img2img or inpaint mode')
    width: int | GraphNode | tuple[GraphNode, str] = Field(default=768, description='Width of output image')
    height: int | GraphNode | tuple[GraphNode, str] = Field(default=768, description='Height of output image')
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='An astronaut riding a rainbow unicorn', description='Input prompt')
    refine: Refine | GraphNode | tuple[GraphNode, str] = Field(default='no_refiner', description='Which refine style to use')
    scheduler: Scheduler | GraphNode | tuple[GraphNode, str] = Field(default='K_EULER', description='scheduler')
    lora_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=0.6, description='LoRA additive scale. Only applicable on trained models.')
    num_outputs: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description='Number of images to output')
    controlnet_1: Controlnet_1 | GraphNode | tuple[GraphNode, str] = Field(default='none', description='Controlnet')
    controlnet_2: Controlnet_2 | GraphNode | tuple[GraphNode, str] = Field(default='none', description='Controlnet')
    controlnet_3: Controlnet_3 | GraphNode | tuple[GraphNode, str] = Field(default='none', description='Controlnet')
    lora_weights: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Replicate LoRA weights to use. Leave blank to use the default weights.')
    refine_steps: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='For base_image_refiner, the number of steps to refine, defaults to num_inference_steps')
    guidance_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=7.5, description='Scale for classifier-free guidance')
    apply_watermark: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='Applies a watermark to enable determining if an image is generated in downstream applications. If you have other provisions for generating or deploying images safely, you can use this to disable watermarking.')
    negative_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='Negative Prompt')
    prompt_strength: float | GraphNode | tuple[GraphNode, str] = Field(default=0.8, description='Prompt strength when using img2img / inpaint. 1.0 corresponds to full destruction of information in image')
    sizing_strategy: Sizing_strategy | GraphNode | tuple[GraphNode, str] = Field(default='width_height', description='Decide how to resize images – use width/height, resize based on input image or control image')
    controlnet_1_end: float | GraphNode | tuple[GraphNode, str] = Field(default=1, description='When controlnet conditioning ends')
    controlnet_2_end: float | GraphNode | tuple[GraphNode, str] = Field(default=1, description='When controlnet conditioning ends')
    controlnet_3_end: float | GraphNode | tuple[GraphNode, str] = Field(default=1, description='When controlnet conditioning ends')
    controlnet_1_image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image for first controlnet')
    controlnet_1_start: float | GraphNode | tuple[GraphNode, str] = Field(default=0, description='When controlnet conditioning starts')
    controlnet_2_image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image for second controlnet')
    controlnet_2_start: float | GraphNode | tuple[GraphNode, str] = Field(default=0, description='When controlnet conditioning starts')
    controlnet_3_image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image for third controlnet')
    controlnet_3_start: float | GraphNode | tuple[GraphNode, str] = Field(default=0, description='When controlnet conditioning starts')
    num_inference_steps: int | GraphNode | tuple[GraphNode, str] = Field(default=30, description='Number of denoising steps')
    disable_safety_checker: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='Disable safety checker for generated images. This feature is only available through the API.')
    controlnet_1_conditioning_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=0.75, description='How strong the controlnet conditioning is')
    controlnet_2_conditioning_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=0.75, description='How strong the controlnet conditioning is')
    controlnet_3_conditioning_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=0.75, description='How strong the controlnet conditioning is')
    @classmethod
    def get_node_type(cls): return "replicate.image.generate.RealVisXLV3"


from nodetool.nodes.replicate.image.generate import Img_size
from nodetool.nodes.replicate.image.generate import Scheduler
from nodetool.nodes.replicate.image.generate import Product_fill

class SDXL_Ad_Inpaint(GraphNode):
    seed: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Empty or 0 for a random image')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Remove background from this image')
    prompt: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Describe the new setting for your product')
    img_size: Img_size | GraphNode | tuple[GraphNode, str] = Field(default='1024, 1024', description='Possible SDXL image sizes')
    apply_img: bool | GraphNode | tuple[GraphNode, str] = Field(default=True, description='Applies the original product image to the final result')
    scheduler: Scheduler | GraphNode | tuple[GraphNode, str] = Field(default='K_EULER', description='scheduler')
    product_fill: Product_fill | GraphNode | tuple[GraphNode, str] = Field(default='Original', description='What percentage of the image width to fill with product')
    guidance_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=7.5, description='Guidance Scale')
    condition_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=0.9, description='controlnet conditioning scale for generalization')
    negative_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='low quality, out of frame, illustration, 3d, sepia, painting, cartoons, sketch, watermark, text, Logo, advertisement', description='Describe what you do not want in your setting')
    num_refine_steps: int | GraphNode | tuple[GraphNode, str] = Field(default=10, description='Number of steps to refine')
    num_inference_steps: int | GraphNode | tuple[GraphNode, str] = Field(default=40, description='Inference Steps')
    @classmethod
    def get_node_type(cls): return "replicate.image.generate.SDXL_Ad_Inpaint"



class SDXL_Controlnet(GraphNode):
    seed: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='Random seed. Set to 0 to randomize the seed')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image for img2img or inpaint mode')
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='aerial view, a futuristic research complex in a bright foggy jungle, hard lighting', description='Input prompt')
    condition_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=0.5, description='controlnet conditioning scale for generalization')
    negative_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='low quality, bad quality, sketches', description='Input Negative Prompt')
    num_inference_steps: int | GraphNode | tuple[GraphNode, str] = Field(default=50, description='Number of denoising steps')
    @classmethod
    def get_node_type(cls): return "replicate.image.generate.SDXL_Controlnet"


from nodetool.nodes.replicate.image.generate import Refine
from nodetool.nodes.replicate.image.generate import Scheduler

class SDXL_Emoji(GraphNode):
    mask: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input mask for inpaint mode. Black areas will be preserved, white areas will be inpainted.')
    seed: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Random seed. Leave blank to randomize the seed')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image for img2img or inpaint mode')
    width: int | GraphNode | tuple[GraphNode, str] = Field(default=1024, description='Width of output image')
    height: int | GraphNode | tuple[GraphNode, str] = Field(default=1024, description='Height of output image')
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='An astronaut riding a rainbow unicorn', description='Input prompt')
    refine: Refine | GraphNode | tuple[GraphNode, str] = Field(default='no_refiner', description='Which refine style to use')
    scheduler: Scheduler | GraphNode | tuple[GraphNode, str] = Field(default='K_EULER', description='scheduler')
    lora_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=0.6, description='LoRA additive scale. Only applicable on trained models.')
    num_outputs: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description='Number of images to output.')
    refine_steps: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='For base_image_refiner, the number of steps to refine, defaults to num_inference_steps')
    guidance_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=7.5, description='Scale for classifier-free guidance')
    apply_watermark: bool | GraphNode | tuple[GraphNode, str] = Field(default=True, description='Applies a watermark to enable determining if an image is generated in downstream applications. If you have other provisions for generating or deploying images safely, you can use this to disable watermarking.')
    high_noise_frac: float | GraphNode | tuple[GraphNode, str] = Field(default=0.8, description='For expert_ensemble_refiner, the fraction of noise to use')
    negative_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='Input Negative Prompt')
    prompt_strength: float | GraphNode | tuple[GraphNode, str] = Field(default=0.8, description='Prompt strength when using img2img / inpaint. 1.0 corresponds to full destruction of information in image')
    replicate_weights: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Replicate LoRA weights to use. Leave blank to use the default weights.')
    num_inference_steps: int | GraphNode | tuple[GraphNode, str] = Field(default=50, description='Number of denoising steps')
    @classmethod
    def get_node_type(cls): return "replicate.image.generate.SDXL_Emoji"


from nodetool.nodes.replicate.image.generate import Refine
from nodetool.nodes.replicate.image.generate import Scheduler

class SDXL_Pixar(GraphNode):
    mask: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input mask for inpaint mode. Black areas will be preserved, white areas will be inpainted.')
    seed: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Random seed. Leave blank to randomize the seed')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image for img2img or inpaint mode')
    width: int | GraphNode | tuple[GraphNode, str] = Field(default=1024, description='Width of output image')
    height: int | GraphNode | tuple[GraphNode, str] = Field(default=1024, description='Height of output image')
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='An astronaut riding a rainbow unicorn', description='Input prompt')
    refine: Refine | GraphNode | tuple[GraphNode, str] = Field(default='no_refiner', description='Which refine style to use')
    scheduler: Scheduler | GraphNode | tuple[GraphNode, str] = Field(default='K_EULER', description='scheduler')
    lora_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=0.6, description='LoRA additive scale. Only applicable on trained models.')
    num_outputs: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description='Number of images to output.')
    refine_steps: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='For base_image_refiner, the number of steps to refine, defaults to num_inference_steps')
    guidance_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=7.5, description='Scale for classifier-free guidance')
    apply_watermark: bool | GraphNode | tuple[GraphNode, str] = Field(default=True, description='Applies a watermark to enable determining if an image is generated in downstream applications. If you have other provisions for generating or deploying images safely, you can use this to disable watermarking.')
    high_noise_frac: float | GraphNode | tuple[GraphNode, str] = Field(default=0.8, description='For expert_ensemble_refiner, the fraction of noise to use')
    negative_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='Input Negative Prompt')
    prompt_strength: float | GraphNode | tuple[GraphNode, str] = Field(default=0.8, description='Prompt strength when using img2img / inpaint. 1.0 corresponds to full destruction of information in image')
    replicate_weights: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Replicate LoRA weights to use. Leave blank to use the default weights.')
    num_inference_steps: int | GraphNode | tuple[GraphNode, str] = Field(default=50, description='Number of denoising steps')
    @classmethod
    def get_node_type(cls): return "replicate.image.generate.SDXL_Pixar"


from nodetool.nodes.replicate.image.generate import Width
from nodetool.nodes.replicate.image.generate import Height
from nodetool.nodes.replicate.image.generate import Scheduler

class StableDiffusion(GraphNode):
    seed: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Random seed. Leave blank to randomize the seed')
    width: Width | GraphNode | tuple[GraphNode, str] = Field(default=768, description='Width of generated image in pixels. Needs to be a multiple of 64')
    height: Height | GraphNode | tuple[GraphNode, str] = Field(default=768, description='Height of generated image in pixels. Needs to be a multiple of 64')
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='a vision of paradise. unreal engine', description='Input prompt')
    scheduler: Scheduler | GraphNode | tuple[GraphNode, str] = Field(default='DPMSolverMultistep', description='Choose a scheduler.')
    num_outputs: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description='Number of images to generate.')
    guidance_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=7.5, description='Scale for classifier-free guidance')
    negative_prompt: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Specify things to not see in the output')
    num_inference_steps: int | GraphNode | tuple[GraphNode, str] = Field(default=50, description='Number of denoising steps')
    @classmethod
    def get_node_type(cls): return "replicate.image.generate.StableDiffusion"


from nodetool.nodes.replicate.image.generate import Scheduler

class StableDiffusionInpainting(GraphNode):
    mask: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description="Mask image - make sure it's the same size as the input image")
    seed: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Random seed. Leave blank to randomize the seed')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image')
    steps: int | GraphNode | tuple[GraphNode, str] = Field(default=20, description='Number of denoising steps')
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='modern bed with beige sheet and pillows', description='Input prompt')
    strength: float | GraphNode | tuple[GraphNode, str] = Field(default=0.7, description='1.0 corresponds to full destruction of information in image')
    scheduler: Scheduler | GraphNode | tuple[GraphNode, str] = Field(default='K_EULER', description='scheduler')
    num_outputs: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description='Number of images to output. Higher number of outputs may OOM.')
    guidance_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=8, description='Guidance scale')
    negative_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='monochrome, lowres, bad anatomy, worst quality, low quality', description='Input Negative Prompt')
    @classmethod
    def get_node_type(cls): return "replicate.image.generate.StableDiffusionInpainting"


from nodetool.nodes.replicate.image.generate import Refine
from nodetool.nodes.replicate.image.generate import Scheduler

class StableDiffusionXL(GraphNode):
    mask: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input mask for inpaint mode. Black areas will be preserved, white areas will be inpainted.')
    seed: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Random seed. Leave blank to randomize the seed')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image for img2img or inpaint mode')
    width: int | GraphNode | tuple[GraphNode, str] = Field(default=1024, description='Width of output image')
    height: int | GraphNode | tuple[GraphNode, str] = Field(default=1024, description='Height of output image')
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='An astronaut riding a rainbow unicorn', description='Input prompt')
    refine: Refine | GraphNode | tuple[GraphNode, str] = Field(default='no_refiner', description='Which refine style to use')
    scheduler: Scheduler | GraphNode | tuple[GraphNode, str] = Field(default='K_EULER', description='scheduler')
    lora_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=0.6, description='LoRA additive scale. Only applicable on trained models.')
    num_outputs: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description='Number of images to output.')
    refine_steps: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='For base_image_refiner, the number of steps to refine, defaults to num_inference_steps')
    guidance_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=7.5, description='Scale for classifier-free guidance')
    apply_watermark: bool | GraphNode | tuple[GraphNode, str] = Field(default=True, description='Applies a watermark to enable determining if an image is generated in downstream applications. If you have other provisions for generating or deploying images safely, you can use this to disable watermarking.')
    high_noise_frac: float | GraphNode | tuple[GraphNode, str] = Field(default=0.8, description='For expert_ensemble_refiner, the fraction of noise to use')
    negative_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='Input Negative Prompt')
    prompt_strength: float | GraphNode | tuple[GraphNode, str] = Field(default=0.8, description='Prompt strength when using img2img / inpaint. 1.0 corresponds to full destruction of information in image')
    replicate_weights: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Replicate LoRA weights to use. Leave blank to use the default weights.')
    num_inference_steps: int | GraphNode | tuple[GraphNode, str] = Field(default=50, description='Number of denoising steps')
    disable_safety_checker: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='Disable safety checker for generated images. This feature is only available through the API. See https://replicate.com/docs/how-does-replicate-work#safety')
    @classmethod
    def get_node_type(cls): return "replicate.image.generate.StableDiffusionXL"


from nodetool.nodes.replicate.image.generate import Scheduler

class StableDiffusionXLLightning(GraphNode):
    seed: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Random seed. Leave blank to randomize the seed')
    width: int | GraphNode | tuple[GraphNode, str] = Field(default=1024, description='Width of output image. Recommended 1024 or 1280')
    height: int | GraphNode | tuple[GraphNode, str] = Field(default=1024, description='Height of output image. Recommended 1024 or 1280')
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='A superhero smiling', description='Input prompt')
    scheduler: Scheduler | GraphNode | tuple[GraphNode, str] = Field(default='K_EULER', description='scheduler')
    num_outputs: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description='Number of images to output.')
    guidance_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=0, description='Scale for classifier-free guidance. Recommended 7-8')
    negative_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='worst quality, low quality', description='Negative Input prompt')
    num_inference_steps: int | GraphNode | tuple[GraphNode, str] = Field(default=4, description='Number of denoising steps. 4 for best results')
    disable_safety_checker: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='Disable safety checker for generated images')
    @classmethod
    def get_node_type(cls): return "replicate.image.generate.StableDiffusionXLLightning"





from datetime import datetime
import json
import os
from typing import Any
from nodetool.api.types.workflow import Workflow
from nodetool.api.types.graph import Graph
from nodetool.workflows.read_graph import read_graph

examples_folder = os.path.join(os.path.dirname(__file__), "examples")


def load_example(name: str) -> Graph:
    with open(os.path.join(examples_folder, f"{name}.json"), "r") as f:
        data = json.load(f)
        edges, nodes = read_graph(data)
        return Graph(nodes=nodes, edges=edges)


def get_examples() -> list[Workflow]:
    return [
        Workflow(
            id="gpt_image",
            access="private",
            created_at="2021-01-01T00:00:00",
            updated_at="2021-01-01T00:00:00",
            name="GPT Image",
            description="A basic workflow for generating images using GPT-3",
            thumbnail_url="/examples/gpt_image.png",
            graph=load_example("gpt_image"),
        ),
        Workflow(
            id="csv_diffusion",
            access="private",
            created_at="2021-01-01T00:00:00",
            updated_at="2021-01-01T00:00:00",
            name="CSV Diffusion",
            description="A basic workflow for inputting CSV data into a diffusion model",
            thumbnail_url="/examples/csv_diffusion.png",
            graph=load_example("csv_diffusion"),
        ),
    ]


def get_comfy_examples() -> list[Workflow]:
    """keep this for comfy launch"""
    return [
        Workflow(
            id="stable_diffusion",
            access="private",
            created_at="2021-01-01T00:00:00",
            updated_at="2021-01-01T00:00:00",
            name="Stable Diffusion",
            description="A basic Comfy workflow for generating photo-realistic images",
            thumbnail_url="/examples/sd.png",
            graph=load_example("stable_diffusion"),
        ),
        Workflow(
            id="realesrgan_x2",
            access="private",
            created_at="2021-01-01T00:00:00",
            updated_at="2021-01-01T00:00:00",
            name="Real-ESRGAN",
            description="Comfy workflow for enhancing images using Real-ESRGAN",
            thumbnail_url="/examples/real_esrgan.jpeg",
            graph=load_example("realesrgan_x2"),
        ),
        Workflow(
            id="lora",
            access="private",
            created_at="2021-01-01T00:00:00",
            updated_at="2021-01-01T00:00:00",
            name="LoRA",
            description="Comfy workflow for applying LORAs to models",
            thumbnail_url="/examples/lora.png",
            graph=load_example("lora"),
        ),
        Workflow(
            id="ipadapter_15",
            access="private",
            created_at="2021-01-01T00:00:00",
            updated_at="2021-01-01T00:00:00",
            name="IPAdapter",
            description="Comfy workflow for applying IPAdapter to models",
            thumbnail_url="/examples/ip_adapter.png",
            graph=load_example("ipadapter_15"),
        ),
    ]



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode

from nodetool.nodes.comfy.sampling import SamplerEnum

class KSamplerSelect(GraphNode):
    sampler_name: SamplerEnum | GraphNode | tuple[GraphNode, str] = Field(default=SamplerEnum('ddim'), description='The name of the sampler.')
    @classmethod
    def get_node_type(cls): return "comfy.sampling.samplers.KSamplerSelect"



class SamplerCustom(GraphNode):
    model: UNet | GraphNode | tuple[GraphNode, str] = Field(default=UNet(type='comfy.unet'), description='The model used by the sampler.')
    add_noise: bool | GraphNode | tuple[GraphNode, str] = Field(default=True, description='Whether to add noise or not.')
    noise_seed: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The seed for the noise generation.')
    cfg: float | GraphNode | tuple[GraphNode, str] = Field(default=8.0, description='The cfg (classifier-free guidance) parameter.')
    positive: Conditioning | GraphNode | tuple[GraphNode, str] = Field(default=Conditioning(type='comfy.conditioning'), description='The positive conditioning.')
    negative: Conditioning | GraphNode | tuple[GraphNode, str] = Field(default=Conditioning(type='comfy.conditioning'), description='The negative conditioning.')
    sampler: Sampler | GraphNode | tuple[GraphNode, str] = Field(default=Sampler(type='comfy.sampler'), description='The sampler to use.')
    sigmas: Sigmas | GraphNode | tuple[GraphNode, str] = Field(default=Sigmas(type='comfy.sigmas'), description='The sigmas used in sampling.')
    latent_image: Latent | GraphNode | tuple[GraphNode, str] = Field(default=Latent(type='comfy.latent'), description='The latent image to sample from.')
    @classmethod
    def get_node_type(cls): return "comfy.sampling.samplers.SamplerCustom"


from nodetool.nodes.comfy.sampling.samplers import SolverTypeEnum
from nodetool.nodes.comfy.sampling.samplers import DeviceEnum

class SamplerDPMPP_2M_SDE(GraphNode):
    solver_type: SolverTypeEnum | GraphNode | tuple[GraphNode, str] = Field(default=SolverTypeEnum('midpoint'), description='The type of solver.')
    eta: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The eta parameter.')
    s_noise: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The scale noise factor.')
    noise_device: DeviceEnum | GraphNode | tuple[GraphNode, str] = Field(default=DeviceEnum('gpu'), description="The device for noise generation, either 'gpu' or 'cpu'.")
    @classmethod
    def get_node_type(cls): return "comfy.sampling.samplers.SamplerDPMPP_2M_SDE"


from nodetool.nodes.comfy.sampling.samplers import DeviceEnum

class SamplerDPMPP_SDE(GraphNode):
    eta: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The eta parameter.')
    s_noise: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The scale noise factor.')
    r: float | GraphNode | tuple[GraphNode, str] = Field(default=0.5, description='The r parameter.')
    noise_device: DeviceEnum | GraphNode | tuple[GraphNode, str] = Field(default=DeviceEnum('gpu'), description="The device for noise generation, either 'gpu' or 'cpu'.")
    @classmethod
    def get_node_type(cls): return "comfy.sampling.samplers.SamplerDPMPP_SDE"





from nodetool.api.types.graph import Graph


from pydantic import BaseModel


from typing import Any, Literal


class RunJobRequest(BaseModel):
    type: Literal["run_job_request"] = "run_job_request"
    job_type: str = "workflow"
    params: Any | None = None
    workflow_id: str = ""
    user_id: str = ""
    auth_token: str = ""
    env: dict[str, Any] | None = None
    graph: Graph | None = None



import json
import re
from jsonpath_ng import parse
from typing import Any
from pydantic import Field
from nodetool.workflows.processing_context import ProcessingContext
from nodetool.metadata.types import LlamaModel, Tensor, TextRef
from nodetool.workflows.base_node import BaseNode
from nodetool.nodes.nodetool.text import convert_result, to_string


class Extract(BaseNode):
    """
    This node extracts a substring from a string.

    #### Inputs
    - `text`: The string from which the substring will be extracted.
    - `start`: The starting index of the substring.
    - `end`: The ending index of the substring.
    """

    text: str | TextRef = Field(title="Text", default="")
    start: int = Field(title="Start", default=0)
    end: int = Field(title="End", default=0)

    async def process(self, context: ProcessingContext) -> str | TextRef:
        text = await to_string(context, self.text)
        res = text[self.start : self.end]
        return await convert_result(context, [self.text], res)


class Chunk(BaseNode):
    """
    This node chunks a string into substrings of a specified number of words.
    """

    text: str | TextRef = Field(title="Text", default="")
    length: int = Field(title="Length", default=100, le=1000, ge=1)
    overlap: int = Field(title="Overlap", default=0)
    separator: str | None = Field(title="Separator", default=None)

    async def process(self, context: ProcessingContext) -> list[str]:
        text = await to_string(context, self.text)
        text = text.split(sep=self.separator)
        chunks = [
            text[i : i + self.length]
            for i in range(0, len(text), self.length - self.overlap)
        ]
        return [" ".join(chunk) for chunk in chunks]


class ExtractRegex(BaseNode):
    """
    This node extracts a list of substrings from a string using a regular expression.
    Each group in the regular expression is extracted as a separate substring.

    #### Applications
    - Extracting substrings from a string using a regular expression.

    #### Example
    Suppose you have a string that contains a question and answer.
    Using a regex you can extract the question and answer as separate substrings.
    """

    text: str | TextRef = Field(title="Text", default="")
    regex: str = Field(title="Regex", default="")
    dotall: bool = Field(title="Dotall", default=False)
    ignorecase: bool = Field(title="Ignorecase", default=False)
    multiline: bool = Field(title="Multiline", default=False)

    async def process(self, context: ProcessingContext) -> list[str]:
        options = 0
        if self.dotall:
            options |= re.DOTALL
        if self.ignorecase:
            options |= re.IGNORECASE
        if self.multiline:
            options |= re.MULTILINE
        text = await to_string(context, self.text)
        match = re.search(self.regex, text, options)
        if match is None:
            return []
        return list(match.groups())


class FindAllRegex(BaseNode):
    """
    This node extracts a list of substrings from a string using a regular expression.
    Each match in the regular expression is extracted as a separate substring.
    """

    text: str | TextRef = Field(title="Text", default="")
    regex: str = Field(title="Regex", default="")
    dotall: bool = Field(title="Dotall", default=False)
    ignorecase: bool = Field(title="Ignorecase", default=False)
    multiline: bool = Field(title="Multiline", default=False)

    async def process(self, context: ProcessingContext) -> list[str]:
        options = 0
        if self.dotall:
            options |= re.DOTALL
        if self.ignorecase:
            options |= re.IGNORECASE
        if self.multiline:
            options |= re.MULTILINE
        text = await to_string(context, self.text)
        matches = re.findall(self.regex, text, options)
        return list(matches)


class ParseJSON(BaseNode):
    """
    This node parses a JSON string into a Python object.

    #### Applications
    - Parsing JSON strings into Python objects for further processing.
    """

    text: str | TextRef = Field(title="JSON string", default="")

    async def process(self, context: ProcessingContext) -> Any:
        json_string = await to_string(context, self.text)
        return json.loads(json_string)


class ExtractJSON(BaseNode):
    """
    This node uses a JSONPath to extract a specific object from a JSON object.

    The Extract JSON Node operates by accepting an input of a JSON object and a JSONPath expression then
    returns the object matching that JSONPath from the JSON object.

    #### Applications
    - Extracting specific data from a larger JSON object as part of a data processing or analysis workflow.
    - Retrieving and analyzing specific values from complex JSON structures.
    - Preprocessing data for subsequent nodes in an AI workflow.

    #### Example
    Suppose you have a large JSON object containing various pieces of information and you want to extract the price information.
    You can use the Extract JSON Node by feeding in the JSON object and specifying the JSONPath expression to the price field.
    The node will return the price information which can then be input to another node for further processing, such as a decision-making node or a data visualization node.
    """

    text: str | TextRef = Field(title="JSON Text", default_factory=TextRef)
    json_path: str = Field(title="JSONPath Expression", default="$.*")
    find_all: bool = Field(title="Find All", default=False)

    async def process(self, context: ProcessingContext) -> Any:
        json_string = await to_string(context, self.text)
        parsed_json = json.loads(json_string)
        jsonpath_expr = parse(self.json_path)
        if self.find_all:
            return [match.value for match in jsonpath_expr.find(parsed_json)]
        else:
            return jsonpath_expr.find(parsed_json)[0].value


class Embedding(BaseNode):
    """
    Generates a vector representation of text for measuring relatedness.
    text, analyse, transform, embeddings, relatedness, search, classification, clustering, recommendations
    Outputs a text embedding vector that quantifies the semantic similarity of the input text to other text strings. An embedding is a vector (list) of floating point numbers. The distance between two vectors measures their relatedness. Small distances suggest high relatedness and large distances suggest low relatedness. Use cases: Search, Clustering, Recommendations, Anomaly detection, Diversity measurement, Classification
    """

    input: str | TextRef = Field(title="Input", default="")
    model: LlamaModel = Field(title="Model", default=LlamaModel())
    n_gpu_layers: int = Field(
        title="Number of GPU Layers",
        default=0,
        ge=-1,
        description="The number of layers on the GPU",
    )
    chunk_size: int = Field(
        title="Chunk Size",
        default=4096,
        ge=64,
        description="The size of the chunks to split the input into",
    )

    async def process(self, context: ProcessingContext) -> Tensor:
        import numpy as np

        model = context.load_llama_model(
            self.model.name, embedding=True, n_gpu_layers=self.n_gpu_layers
        )
        input = await context.to_str(self.input)
        # chunk the input into smaller pieces
        chunks = [
            input[i : i + self.chunk_size]
            for i in range(0, len(input), self.chunk_size)
        ]
        res = model.create_embedding(input=chunks)
        all = [i["embedding"] for i in res["data"]]
        avg = np.mean(all, axis=0)
        return Tensor.from_numpy(avg)



import numpy as np
from nodetool.metadata.types import Tensor
from nodetool.workflows.processing_context import ProcessingContext
from nodetool.nodes.nodetool.audio.audio_helpers import (
    numpy_to_audio_segment,
)
from nodetool.metadata.types import AudioRef
from nodetool.workflows.base_node import BaseNode
from pydantic import Field


class AudioToTensor(BaseNode):
    """
    AudioToTensor node transforms an audio file into a tensor, a data structure that can be understood by machine learning algorithms.

    This node is primarily useful for audio data pre-processing, as machine learning algorithms require data in a specific format (like tensors). The resultant tensor contains the raw audio samples ready for further processing. Some of the significant uses of this node are in voice recognition, sound classification, and other audio analysis tasks.

    Applications:
    - Transforming audio files for machine learning tasks.
    - Converting audio into a manageable and processable format for further audio analysis.
    """

    audio: AudioRef = Field(
        default=AudioRef(), description="The audio file to convert to a tensor."
    )

    async def process(self, context: ProcessingContext) -> Tensor:
        audio = await context.audio_to_audio_segment(self.audio)
        samples = np.array(audio.get_array_of_samples().tolist())
        return Tensor.from_numpy(samples)


class TensorToAudio(BaseNode):
    """
    This node converts a numerical tensor object to an audio file.

    A tensor is a mathematical object analogous to but more general than a vector, represented by an array of components that are functions of the coordinates of a space. In this case, it represents the audio data. This node takes such tensor data and transforms it into an audio format.

    The node is handy when dealing with audio data or signals that have been transformed into tensor format for machine learning models or signal processing. The node allows you to convert the tensor back to an audio file. This is particularly essential when you need to listen to the model output or the processed signals.

    Applications:
    - Reverse the output of machine learning models that generate tensor audio data back to an audiofile.
    - Retrieve and listen to the audio signal after audio signal processing done on tensor format.
    """

    tensor: Tensor = Field(
        default=Tensor(), description="The tensor to convert to an audio file."
    )
    sample_rate: int = Field(
        default=44100, ge=0, le=44100, description="The sample rate of the audio file."
    )

    async def process(self, context: ProcessingContext) -> AudioRef:
        audio = numpy_to_audio_segment(self.tensor.to_numpy(), self.sample_rate)
        return await context.audio_from_segment(audio)



from time import sleep
from typing import Any, Type

import os
import httpx
import asyncio
import re
from bs4 import BeautifulSoup

from datetime import datetime
from typing import Any, Type, get_origin
from openapi_pydantic.v3 import (
    parse_obj,
    DataType,
    Schema,
    Reference,
)
from openapi_pydantic.v3.parser import OpenAPIv3
from enum import Enum

from pydantic import ConfigDict
from nodetool.common.environment import Environment
from nodetool.dsl.codegen import field_default, type_to_string
from nodetool.metadata.types import Tensor
from nodetool.metadata.utils import is_enum_type
from nodetool.models.prediction import Prediction
from nodetool.workflows.processing_context import ProcessingContext
from nodetool.metadata.types import AudioRef
from nodetool.metadata.types import ImageRef
from nodetool.metadata.types import VideoRef
from nodetool.workflows.types import NodeUpdate
from nodetool.workflows.base_node import BaseNode
from nodetool.workflows.processing_context import ProcessingContext
from replicate.prediction import Prediction as ReplicatePrediction
import httpx
import asyncio
from typing import Any, Type
import os


log = Environment.get_logger()
current_folder = os.path.dirname(os.path.abspath(__file__))
replicate_nodes_folder = os.path.join(
    os.path.dirname(current_folder), "nodes", "replicate"
)

REPLICATE_MODELS = {}


def add_replicate_model(model_id: str, model_info: dict[str, Any]):
    REPLICATE_MODELS[model_id] = model_info


class ReplicateNode(BaseNode):
    """
    This is the base class for all replicate nodes.

    Methods:
        replicate_model_id: Returns the model ID for replication.
        get_hardware: Returns the hardware information.
        run_replicate: Runs prediction on Replicate.
        extra_params: Returns any extra parameters.
        process: Processes the prediction output.
        convert_output: Converts the output to the specified type.

    """

    model_config = ConfigDict(protected_namespaces=())

    @classmethod
    def is_visible(cls) -> bool:
        return cls is not ReplicateNode

    @classmethod
    def __init_subclass__(cls):
        """
        This method is called when a subclass of this class is created.
        We remember the model_id and corresponding model_info for each subclass.
        """
        super().__init_subclass__()
        model_info = cls.model_info().copy()
        model_info["hardware"] = cls.get_hardware()
        add_replicate_model(cls.replicate_model_id(), model_info)

    @classmethod
    def replicate_model_id(cls) -> str:
        """
        Subclasses need to implement this method to return the model ID to run.
        """
        raise NotImplementedError()

    @classmethod
    def get_hardware(cls) -> str:
        """
        Subclasses need to implement this method to return the hardware information.
        """
        return ""

    def get_output_index(self) -> int:
        return 0

    def output_key(self) -> str:
        return "output"

    async def run_replicate(
        self, context: ProcessingContext, params: dict[(str, Any)] | None = None
    ):
        """
        Run prediction on Replicate.

        Args:
            context: The processing context.
            params: Optional dictionary of parameters.

        Returns:
            Result of the prediction.
        """
        raw_inputs = {
            prop.name: convert_enum_value(getattr(self, prop.name))
            for prop in self.properties()
        }
        raw_inputs = {**raw_inputs, **(params or {})}

        input_params = {
            prop.name: await context.convert_value_for_prediction(
                prop,
                getattr(self, prop.name),
            )
            for prop in self.properties()
        }
        input_params = {
            key: value for (key, value) in input_params.items() if (value is not None)
        }
        input_params = {**input_params, **(params or {})}

        res = await context.run_prediction(
            provider="replicate",
            node_id=self.id,
            model=self.replicate_model_id(),
            params=input_params,
        )
        if res.media_type == "application/json":
            return res.json()
        else:
            return res.content

    async def process(self, context: ProcessingContext):
        """
        Process the prediction output.

        Args:
            context: The processing context.

        Returns:
            The processed output.

        Raises:
            ValueError: If prediction failed.

        """
        result = await self.run_replicate(context)
        return result

    async def convert_output(self, context: ProcessingContext, output: Any) -> Any:
        """
        Convert the output to the specified type.

        Args:
            context: The processing context.
            output: The output to be converted.

        Returns:
            The converted output.

        """
        t = self.return_type()
        if isinstance(t, dict):
            return {
                key: await convert_output_value(
                    output[key],
                    t[key],
                )
                for key in t
            }
        if isinstance(t, Type):
            output = await convert_output_value(
                output,
                t,
                output_index=self.get_output_index(),
                output_key=self.output_key(),
            )
        return {
            "output": output,
        }


class TypeName:
    """
    A class representing a type name.

    Attributes:
        __name__ (str): The name of the type.
    """

    def __init__(self, name: str):
        self.__name__ = name


def calculate_llm_cost(model: str, input_tokens: int, output_tokens: int) -> float:
    pricing = {
        "meta/meta-llama-3-8b": (0.05, 0.25),
        "meta/meta-llama-3-13b": (0.10, 0.50),
        "meta/meta-llama-3-70b": (0.65, 2.75),
        "meta/meta-llama-3-8b-instruct": (0.05, 0.25),
        "meta/meta-llama-3-13b-instruct": (0.10, 0.50),
        "meta/meta-llama-3-70b-instruct": (0.65, 2.75),
        "mistralai/mistral-7b-v0.2": (0.05, 0.25),
        "mistralai/mistral-7b-instruct-v0.2": (0.05, 0.25),
        "mistralai/mixtral-8x7b-instruct-v0.1": (0.30, 1.00),
        "snowflake/snowflake-arctic-instruct": (20.00, 20.00),
    }

    model = model.split(":")[0]

    if model not in pricing:
        raise ValueError(f"Unsupported model: {model}")

    input_cost_per_million, output_cost_per_million = pricing[model]

    input_cost = input_tokens * input_cost_per_million / 1_000_000
    output_cost = output_tokens * output_cost_per_million / 1_000_000

    return input_cost + output_cost


def calculate_cost(hardware: str, duration: float):
    pricing = {
        "CPU": 0.000100,
        "Nvidia T4 GPU": 0.000225,
        "Nvidia T4 (High-memory) GPU": 0.000225,
        "Nvidia A40 GPU": 0.000575,
        "Nvidia A40 (Large) GPU": 0.000725,
        "Nvidia A100 (40GB) GPU": 0.001150,
        "Nvidia A100 (80GB) GPU": 0.001400,
        "8x Nvidia A40 (Large) GPU": 0.005800,
    }

    if hardware in pricing:
        price_per_second = pricing[hardware]
        total_price = price_per_second * duration
        return total_price
    else:
        raise ValueError(f"Unsupported hardware: {hardware}")


REPLICATE_STATUS_MAP = {
    "starting": "starting",
    "succeeded": "completed",
    "processing": "running",
    "failed": "failed",
    "canceled": "canceled",
}


async def run_replicate(
    prediction: Prediction,
    params: dict,
):
    """
    Run the model on Replicate API
    """
    model = prediction.model
    model_info = REPLICATE_MODELS.get(model) or {}
    hardware = model_info.get("hardware", "")

    replicate = Environment.get_replicate_client()

    log.info(f"Running model {model}")
    started_at = datetime.now()

    # Split model_version into owner, name, version in format owner/name:version
    match = re.match(r"^(?P<owner>[^/]+)/(?P<name>[^:]+):(?P<version>.+)$", model)
    if not match:
        raise ValueError(f"Invalid model: {model}. Expected format: owner/name:version")
    owner = match.group("owner")
    name = match.group("name")
    version_id = match.group("version")
    replicate_pred = replicate.predictions.create(
        version=version_id,
        input=params,
    )

    current_status = "starting"

    for i in range(1800):
        replicate_pred = replicate.predictions.get(replicate_pred.id)

        if replicate_pred.status != current_status:
            log.info(f"Prediction status: {replicate_pred.status}")
            current_status = replicate_pred.status

        prediction.status = REPLICATE_STATUS_MAP[replicate_pred.status]
        prediction.logs = replicate_pred.logs
        prediction.error = replicate_pred.error
        prediction.duration = (datetime.now() - started_at).total_seconds()
        prediction.save()

        if replicate_pred.status in ("failed", "canceled"):
            raise ValueError(replicate_pred.error or "Prediction failed")

        if replicate_pred.status in ("succeeded", "failed", "canceled"):
            break

        await asyncio.sleep(1)

    assert replicate_pred.metrics, "Prediction metrics not found"

    if "input_token_count" in replicate_pred.metrics:
        input_token_count = replicate_pred.metrics["input_token_count"]
        output_token_count = replicate_pred.metrics["output_token_count"]
        prediction.cost = calculate_llm_cost(
            model, input_token_count, output_token_count
        )
        prediction.save()

    elif "predict_time" in replicate_pred.metrics:
        predict_time = replicate_pred.metrics["predict_time"]
        prediction.cost = calculate_cost(hardware, predict_time)
        prediction.save()
    else:
        raise ValueError("Cost calculation failed")

    return replicate_pred.output


async def convert_output_value(
    value: Any, t: Type[Any], output_index: int = 0, output_key: str = "output"
):
    """
    Converts the output value to the specified type.
    Performs automatic conversions using heuristics.

    Args:
        value (Any): The value to be converted.
        t (Type[Any]): The target type to convert the value to.
        output_index: The index for list outputs to use.

    Returns:
        Any: The converted value.

    Raises:
        TypeError: If the value is not of the expected type.

    """
    if value is None:
        return None

    if isinstance(value, t):
        return value

    if t in (ImageRef, AudioRef, VideoRef):
        if type(value) == list:
            return t(uri=str(value[output_index])).model_dump()
        elif type(value) == dict:
            if output_key in value:
                return t(uri=value[output_key]).model_dump()
            else:
                raise ValueError(f"Output key not found: {output_key}")
        elif type(value) == str:
            return t(uri=value).model_dump()
        else:
            raise TypeError(f"Invalid {t} value: {value}")
    elif t == str:
        if type(value) == list:
            return "".join(str(i) for i in value)
        else:
            return str(value)
    elif t == Tensor:
        return Tensor.from_list(value).model_dump()
    elif get_origin(t) == list:
        if type(value) != list:
            raise TypeError(f"value is not list: {value}")
        tasks = [convert_output_value(item, t.__args__[0]) for item in value]  # type: ignore
        return await asyncio.gather(*tasks)
    return value


def convert_enum_value(value: Any):
    """
    Converts an enum value to its corresponding value.

    Args:
        value (Any): The value to be converted.

    Returns:
        Any: The converted value.
    """
    if isinstance(value, Enum):
        return value.value
    return value


def sanitize_enum(name: str) -> str:
    """
    Sanitizes an enum string by replacing hyphens, dots, and spaces with underscores.

    Args:
        enum (str): The enum string to be sanitized.

    Returns:
        str: The sanitized enum string.
    """
    name = re.sub(r"[^a-zA-Z0-9_]", "_", name).upper()
    while len(name) > 0 and name[0] == "_":
        name = name[1:]
    while len(name) > 0 and name[-1] == "_":
        name = name[:-1]
    if name[0].isdigit():
        name = "_" + name
    return name


def capitalize(name: str) -> str:
    """
    Capitalizes the first letter of a string.

    Args:
        name (str): The string to be capitalized.

    Returns:
        str: The capitalized string.
    """
    return name[0].upper() + name[1:]


def create_enum_from_schema(name: str, schema: Schema):
    """
    Create source code for an enum model.

    Args:
        name (str): The name of the enum model.
        schema (Schema): The OpenAPI schema containing the enum field.

    Returns:
        str: The source code of the enum model.

    Raises:
        ValueError: If the schema has no enum field.
    """
    if schema.enum is not None:
        enum_type = type(schema.enum[0])
        values = "".join(
            [
                f"    {sanitize_enum(str(value))} = {repr(value)}\n"
                for value in schema.enum
            ]
        )
        return f"class {name}({enum_type.__name__}, Enum):\n" + values
    else:
        raise ValueError("Schema has no enum field")


def convert_datatype_to_type(
    datatype: DataType | list[DataType], is_optional: bool
) -> type:
    """
    Converts an OpenAPI property schema to a Pydantic field.

    Args:
        datatype (DataType | list[DataType]): The datatype or list of datatypes to convert.
        is_optional (bool): Whether the field is optional.

    Returns:
        type: The type of the field.

    Raises:
        ValueError: If the datatype is a list.

    """
    if isinstance(datatype, list):
        raise ValueError("Cannot convert list of datatypes to type")
    t = None
    if datatype == DataType.STRING:
        t = str
    elif datatype == DataType.NUMBER:
        t = float
    elif datatype == DataType.INTEGER:
        t = int
    elif datatype == DataType.BOOLEAN:
        t = bool
    elif datatype == DataType.ARRAY:
        t = list
    elif datatype == DataType.OBJECT:
        t = dict
    else:
        raise ValueError(f"Unknown type: {datatype}")
    if is_optional:
        return t | None  # type: ignore
    return t


def default_for(datatype: DataType | list[DataType] | None) -> Any:
    if datatype == DataType.STRING:
        return ""
    elif datatype == DataType.NUMBER:
        return 0.0
    elif datatype == DataType.INTEGER:
        return 0
    elif datatype == DataType.BOOLEAN:
        return False
    elif datatype == DataType.ARRAY:
        return []
    elif datatype == DataType.OBJECT:
        return {}
    else:
        return None


def return_type_repr(return_type: type) -> str:
    if isinstance(return_type, dict):
        return (
            "{"
            + ", ".join(f"'{k}': {type_to_string(v)}" for k, v in return_type.items())
            + "}\n"
        )
    else:
        return type_to_string(return_type)


def generate_model_source_code(
    model_name: str,
    model_info: dict[str, Any],
    description: str,
    schema: Schema,
    type_lookup: dict[str, type],
    return_type: type,
    hardware: str,
    replicate_model_id: str,
    overrides: dict[str, type] = {},
    output_index: int = 0,
    output_key: str = "output",
) -> str:
    """
    Generate source code for a Pydantic model from an OpenAPI schema.

    Args:
        model_name (str): The class name of the model.
        model_info (dict[str, Any]): The model information.
        description (str): Description of the model.
        schema (Schema): The schema object.
        type_lookup (Dict[str, Type]): Mapping of type names to Python types.
        return_type (Type): The return type of the model.
        hardware (str): The hardware information.
        replicate_model_id (str): The replicate model ID.
        overrides (Dict[str, Type], optional): Field type overrides. Defaults to {}.
        output_index (int, optional): The index for list outputs to use. Defaults to 0.
        output_key (str, optional): The key for dict outputs to use. Defaults to "output".

    Returns:
        str: The source code of the model.

    Raises:
        ValueError: If the schema has no properties or if the 'allOf' type cannot be handled.
    """

    if not schema.properties:
        raise ValueError("Schema has no properties")

    imports = ""

    lines = [
        f"class {model_name}(ReplicateNode):",
        '    """' + description + '"""',
        "",
        "    @classmethod",
        f"    def replicate_model_id(cls): return '{replicate_model_id}'",
        "    @classmethod",
        f"    def get_hardware(cls): return '{hardware}'",
        "    @classmethod",
        f"    def model_info(cls): return {repr(model_info)}",
    ]
    if return_type:
        lines += [
            "    @classmethod",
            f"    def return_type(cls): return {return_type_repr(return_type)}",
        ]

    if output_index > 0:
        lines.append(f"   def output_index(self): return {output_index}")

    if output_key != "output":
        lines.append(f"    def output_key(self): return '{output_key}'")

    lines.append("\n")

    for name, prop in schema.properties.items():
        if name == "api_key" or isinstance(prop, Reference):
            continue

        is_optional = prop.default is None

        field_type, field_args = Any, ""
        if name in overrides:
            field_type = overrides[name]
            field_args = f" = Field(default={field_type.__name__}(), description={repr(prop.description)})"
        else:
            if prop.type:
                field_type = convert_datatype_to_type(prop.type, is_optional)
            elif prop.allOf:
                ref_type = prop.allOf[0]
                if hasattr(ref_type, "ref"):
                    ref = ref_type.ref  # type: ignore
                    field_type = type_lookup[ref]

            field_args = " = Field("
            if prop.title:
                field_args += f"title='{prop.title}', "
            if prop.description:
                field_args += f"description={repr(prop.description)}, "
            if prop.minimum is not None:
                field_args += f"ge={prop.minimum}, "
            if prop.maximum is not None:
                field_args += f"le={prop.maximum}, "
            field_args += f"default={field_default(prop.default)})"

        if is_enum_type(field_type):
            imports += f"from {field_type.__module__} import {field_type.__name__}\n"

        lines.append(f"    {name}: {type_to_string(field_type)}{field_args}")

    return imports + "\n\n" + "\n".join(lines)


def parse_model_info(url: str):
    """
    Parses the replicate model information from the given URL.

    Args:
        url (str): The URL to fetch the HTML content from.

    Returns:
        dict: A dictionary containing the parsed model information.
    """

    # Fetch the HTML content
    response = httpx.get(url)
    response.raise_for_status()
    html_content = response.content

    # Parse the HTML content with BeautifulSoup
    soup = BeautifulSoup(html_content, "html.parser")

    # Find the section titled "Run time and cost"
    runtime_section = soup.find("div", {"id": "performance"})

    if runtime_section:
        # Extract the hardware information
        hardware_info = runtime_section.find("p", string=lambda text: "hardware" in text.lower())  # type: ignore
        if hardware_info:
            match = re.search(
                r"This model runs on (.+?) hardware", hardware_info.text.strip()
            )
            if match:
                hardware = match.group(1)
            else:
                hardware = hardware_info.text.strip()
        else:
            hardware = "None"

        return {"hardware": hardware}
    else:
        return {}


def retry_http_request(url: str, headers: dict[str, str], retries: int = 3) -> Any:
    """
    Retry an HTTP request with exponential backoff.

    Args:
        url (str): The URL to make the request to.
        headers (dict[str, str]): The headers to include in the request.
        retries (int, optional): The number of retries. Defaults to 3.

    Returns:
        Any: The response data.

    Raises:
        httpx.RequestError: If the request fails after all retries.
    """
    for attempt in range(retries + 1):
        try:
            response = httpx.get(url, headers=headers)
            response.raise_for_status()
            return response.json()
        except (httpx.RequestError, httpx.HTTPStatusError) as e:
            if attempt < retries:
                delay = 2**attempt
                sleep(delay)
            else:
                raise e


def get_model_api(
    model_id: str,
) -> tuple[OpenAPIv3, str, str, dict[str, Any]]:
    """
    Retrieves the API specification for a specific model and version from the Replicate API.

    Args:
        model_id (str): The replicate model ID.
        model_version (str, optional): The model version. Defaults to None.
    """
    model_name = model_id.replace("/", "_").replace("-", "_").replace(".", "_")

    if "REPLICATE_API_TOKEN" not in os.environ:
        raise ValueError("REPLICATE_API_TOKEN environment variable is not set")

    headers = {
        "Authorization": "Token " + Environment.get_replicate_api_token(),
    }

    log.info("Getting model API: %s", model_id)
    model_info = retry_http_request(
        f"https://api.replicate.com/v1/models/{model_id}",
        headers=headers,
    )
    model_info.update(parse_model_info(f"https://replicate.com/{model_id}"))
    latest_version = model_info["latest_version"]
    api = parse_obj(latest_version["openapi_schema"])
    model_version = latest_version.get("id", "")

    return api, model_id, model_version, model_info


classes_by_namespace: dict[str, set] = {}
enums_by_namespace: dict[str, set] = {}


def replicate_node(
    node_name: str,
    namespace: str,
    model_id: str,
    return_type: Type,
    overrides: dict[str, type] = {},
    output_index: int = 0,
    output_key: str = "output",
):
    """
    Creates a node from a model ID and version.
    Gets the model from the replicate API and creates a node class from it.
    The node class is then added to the NODE_TYPES list.

    Args:
        node_name (str): The name of the node.
        namespace (str): The namespace of the node.
        model_id (str): The ID of the model on replicate.
        return_type (Type): The return type of the node.
        model_version (str, optional): The version of the model. Defaults to None.
        overrides (dict[str, type]): A dictionary of overrides for the node fields.
        output_index (int): The index for list outputs to use. Defaults to 0.
        output_key (str): The key for dict outputs to use. Defaults to "output".

    Returns:
        Replicate: The generated node class.

    Raises:
        ValueError: If the schema has no properties or enum.
    """

    type_lookup = {}
    source_code = ""
    if not namespace in enums_by_namespace:
        enums_by_namespace[namespace] = set()

    api, model_id, model_version, model_info = get_model_api(model_id)

    assert api.components
    assert api.components.schemas

    for name, schema in api.components.schemas.items():
        if name in (
            "Input",
            "Output",
            "Request",
            "Response",
            "PredictionRequest",
            "PredictionResponse",
            "Status",
            "WebhookEvent",
            "ValidationError",
            "HTTPValidationError",
        ):
            continue

        if hasattr(schema, "enum") and schema.enum is not None:  # type: ignore
            capitalized_name = capitalize(name)
            type_lookup[f"#/components/schemas/{name}"] = TypeName(capitalized_name)
            if not name in enums_by_namespace[namespace]:
                source_code += create_enum_from_schema(capitalized_name, schema)  # type: ignore
                enums_by_namespace[namespace].add(name)

    del model_info["default_example"]
    del model_info["latest_version"]

    source_code += generate_model_source_code(
        model_name=node_name,
        model_info=model_info,
        description=model_info.get("description", ""),
        schema=api.components.schemas["Input"],  # type: ignore
        type_lookup=type_lookup,
        overrides=overrides,
        return_type=return_type,
        output_index=output_index,
        output_key=output_key,
        hardware=model_info.get("hardware", None),
        replicate_model_id=f"{model_id}:{model_version}",
    )
    namespace_path = os.path.join(replicate_nodes_folder, namespace.replace(".", "/"))
    namespace_folder = os.path.dirname(namespace_path)
    os.makedirs(namespace_folder, exist_ok=True)

    if not namespace in classes_by_namespace:
        classes_by_namespace[namespace] = set()
        imports = (
            "from pydantic import BaseModel, Field\n"
            "from nodetool.metadata.types import *\n"
            "from nodetool.dsl.graph import GraphNode\n"
            "from nodetool.common.replicate_node import ReplicateNode\n"
            "from enum import Enum\n"
        )
        with open(namespace_path + ".py", "w") as f:
            f.write(imports)

    classes_by_namespace[namespace].add(node_name)

    with open(namespace_path + ".py", "a") as f:
        f.write("\n\n")
        f.write(source_code)



from pydantic import BaseModel, Field
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode
from nodetool.common.replicate_node import ReplicateNode
from enum import Enum


class Mode(str, Enum):
    BEST = "best"
    FAST = "fast"


class SDXLClipInterrogator(ReplicateNode):
    """CLIP Interrogator for SDXL optimizes text prompts to match a given image"""

    @classmethod
    def replicate_model_id(cls):
        return "lucataco/sdxl-clip-interrogator:d90ed1292165dbad1fc3fc8ce26c3a695d6a211de00e2bb5f5fec4815ea30e4c"

    @classmethod
    def get_hardware(cls):
        return "Nvidia A40 GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://tjzk.replicate.delivery/models_models_cover_image/c200f919-4cc1-412b-8edf-e2863a5eef56/replicate-sdxl-inter.png",
            "created_at": "2023-08-14T20:06:38.402771Z",
            "description": "CLIP Interrogator for SDXL optimizes text prompts to match a given image",
            "github_url": "https://github.com/lucataco/cog-sdxl-clip-interrogator",
            "license_url": "https://github.com/pharmapsychotic/clip-interrogator/blob/main/LICENSE",
            "name": "sdxl-clip-interrogator",
            "owner": "lucataco",
            "paper_url": None,
            "run_count": 835178,
            "url": "https://replicate.com/lucataco/sdxl-clip-interrogator",
            "visibility": "public",
            "hardware": "Nvidia A40 GPU",
        }

    @classmethod
    def return_type(cls):
        return str

    mode: Mode = Field(
        description="Prompt Mode: fast takes 1-2 seconds, best takes 15-25 seconds.",
        default="best",
    )
    image: ImageRef = Field(default=ImageRef(), description="Input image")


class Moondream2(ReplicateNode):
    """moondream2 is a small vision language model designed to run efficiently on edge devices"""

    @classmethod
    def replicate_model_id(cls):
        return "lucataco/moondream2:392a53ac3f36d630d2d07ce0e78142acaccc338d6caeeb8ca552fe5baca2781e"

    @classmethod
    def get_hardware(cls):
        return "Nvidia A40 GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://tjzk.replicate.delivery/models_models_cover_image/3cbb4e68-08b8-4e82-8e83-3300f877dd0f/moondream2.png",
            "created_at": "2024-03-05T02:29:40.377800Z",
            "description": "moondream2 is a small vision language model designed to run efficiently on edge devices",
            "github_url": "https://github.com/lucataco/cog-moondream2",
            "license_url": "https://github.com/vikhyat/moondream?tab=Apache-2.0-1-ov-file#readme",
            "name": "moondream2",
            "owner": "lucataco",
            "paper_url": None,
            "run_count": 13929,
            "url": "https://replicate.com/lucataco/moondream2",
            "visibility": "public",
            "hardware": "Nvidia A40 GPU",
        }

    @classmethod
    def return_type(cls):
        return str

    image: ImageRef = Field(default=ImageRef(), description="Input image")
    prompt: str = Field(
        title="Prompt", description="Input prompt", default="Describe this image"
    )


class MiniGPT4(ReplicateNode):
    """A model which generates text in response to an input image and prompt."""

    @classmethod
    def replicate_model_id(cls):
        return "daanelson/minigpt-4:b96a2f33cc8e4b0aa23eacfce731b9c41a7d9466d9ed4e167375587b54db9423"

    @classmethod
    def get_hardware(cls):
        return "Nvidia A100 (40GB) GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://tjzk.replicate.delivery/models_models_cover_image/af717919-83de-46e8-9b1a-9c66f4f747bf/out_0.png",
            "created_at": "2023-05-16T19:05:24.691944Z",
            "description": "A model which generates text in response to an input image and prompt.",
            "github_url": "https://github.com/daanelson/MiniGPT-4",
            "license_url": "https://github.com/Vision-CAIR/MiniGPT-4/blob/main/LICENSE.md",
            "name": "minigpt-4",
            "owner": "daanelson",
            "paper_url": "https://arxiv.org/pdf/2304.10592.pdf",
            "run_count": 1341037,
            "url": "https://replicate.com/daanelson/minigpt-4",
            "visibility": "public",
            "hardware": "Nvidia A100 (40GB) GPU",
        }

    @classmethod
    def return_type(cls):
        return str

    image: str | None = Field(
        title="Image", description="Image to discuss", default=None
    )
    top_p: float = Field(
        title="Top P",
        description="Sample from the top p percent most likely tokens",
        ge=0.0,
        le=1.0,
        default=0.9,
    )
    prompt: str | None = Field(
        title="Prompt",
        description="Prompt for mini-gpt4 regarding input image",
        default=None,
    )
    num_beams: int = Field(
        title="Num Beams",
        description="Number of beams for beam search decoding",
        ge=1.0,
        le=10.0,
        default=3,
    )
    max_length: int = Field(
        title="Max Length",
        description="Total length of prompt and output in tokens",
        ge=1.0,
        default=4000,
    )
    temperature: float = Field(
        title="Temperature",
        description="Temperature for generating tokens, lower = more predictable results",
        ge=0.01,
        le=2.0,
        default=1,
    )
    max_new_tokens: int = Field(
        title="Max New Tokens",
        description="Maximum number of new tokens to generate",
        ge=1.0,
        default=3000,
    )
    repetition_penalty: float = Field(
        title="Repetition Penalty",
        description="Penalty for repeated words in generated text; 1 is no penalty, values greater than 1 discourage repetition, less than 1 encourage it.",
        ge=0.01,
        le=5.0,
        default=1,
    )


class NSFWImageDetection(ReplicateNode):
    """Falcons.ai Fine-Tuned Vision Transformer (ViT) for NSFW Image Classification"""

    @classmethod
    def replicate_model_id(cls):
        return "lucataco/nsfw_image_detection:97116600cabd3037e5f22ca08ffcc33b92cfacebf7ccd3609e9c1d29e43d3a8d"

    @classmethod
    def get_hardware(cls):
        return "Nvidia T4 GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://replicate.delivery/pbxt/JurYNQcIfISvpS6WtaOcwZXw1ifEudlLyQqiLj5N1Zq977Q3/falcon.jpg",
            "created_at": "2023-11-21T14:53:34.798862Z",
            "description": "Falcons.ai Fine-Tuned Vision Transformer (ViT) for NSFW Image Classification",
            "github_url": "https://github.com/lucataco/cog-nsfw_image_detection",
            "license_url": "https://huggingface.co/models?license=license:apache-2.0",
            "name": "nsfw_image_detection",
            "owner": "lucataco",
            "paper_url": "https://huggingface.co/papers/2010.11929",
            "run_count": 1465305,
            "url": "https://replicate.com/lucataco/nsfw_image_detection",
            "visibility": "public",
            "hardware": "Nvidia T4 GPU",
        }

    @classmethod
    def return_type(cls):
        return str

    image: str | None = Field(title="Image", description="Input image", default=None)


class Llava34B(ReplicateNode):
    """LLaVA v1.6: Large Language and Vision Assistant (Nous-Hermes-2-34B)"""

    @classmethod
    def replicate_model_id(cls):
        return "yorickvp/llava-v1.6-34b:41ecfbfb261e6c1adf3ad896c9066ca98346996d7c4045c5bc944a79d430f174"

    @classmethod
    def get_hardware(cls):
        return "Nvidia A100 (80GB) GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://tjzk.replicate.delivery/models_models_cover_image/c6163ba0-edfc-4b53-9a23-eab7fd08b28a/b14df1cd-2e49-4e6b-b965-0deea7c1.webp",
            "created_at": "2024-02-01T14:41:55.242062Z",
            "description": "LLaVA v1.6: Large Language and Vision Assistant (Nous-Hermes-2-34B)",
            "github_url": "https://github.com/haotian-liu/LLaVA",
            "license_url": "https://huggingface.co/NousResearch/Nous-Hermes-2-Yi-34B",
            "name": "llava-v1.6-34b",
            "owner": "yorickvp",
            "paper_url": None,
            "run_count": 1076540,
            "url": "https://replicate.com/yorickvp/llava-v1.6-34b",
            "visibility": "public",
            "hardware": "Nvidia A100 (80GB) GPU",
        }

    @classmethod
    def return_type(cls):
        return str

    image: ImageRef = Field(default=ImageRef(), description="Input image")
    top_p: float = Field(
        title="Top P",
        description="When decoding text, samples from the top p percentage of most likely tokens; lower to ignore less likely tokens",
        ge=0.0,
        le=1.0,
        default=1,
    )
    prompt: str | None = Field(
        title="Prompt", description="Prompt to use for text generation", default=None
    )
    history: list | None = Field(
        title="History",
        description="List of earlier chat messages, alternating roles, starting with user input. Include <image> to specify which message to attach the image to.",
        default=None,
    )
    max_tokens: int = Field(
        title="Max Tokens",
        description="Maximum number of tokens to generate. A word is generally 2-3 tokens",
        ge=0.0,
        default=1024,
    )
    temperature: float = Field(
        title="Temperature",
        description="Adjusts randomness of outputs, greater than 1 is random and 0 is deterministic",
        ge=0.0,
        default=0.2,
    )


class Task(str, Enum):
    IMAGE_CAPTIONING = "image_captioning"
    VISUAL_QUESTION_ANSWERING = "visual_question_answering"
    IMAGE_TEXT_MATCHING = "image_text_matching"


class Blip(ReplicateNode):
    """Bootstrapping Language-Image Pre-training"""

    @classmethod
    def replicate_model_id(cls):
        return "salesforce/blip:2e1dddc8621f72155f24cf2e0adbde548458d3cab9f00c0139eea840d0ac4746"

    @classmethod
    def get_hardware(cls):
        return "Nvidia T4 GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://replicate.delivery/mgxm/f4e50a7b-e8ca-432f-8e68-082034ebcc70/demo.jpg",
            "created_at": "2022-02-06T17:40:38.855280Z",
            "description": "Bootstrapping Language-Image Pre-training",
            "github_url": "https://github.com/salesforce/BLIP",
            "license_url": "https://github.com/salesforce/BLIP/blob/main/LICENSE.txt",
            "name": "blip",
            "owner": "salesforce",
            "paper_url": "https://arxiv.org/abs/2201.12086",
            "run_count": 78252387,
            "url": "https://replicate.com/salesforce/blip",
            "visibility": "public",
            "hardware": "Nvidia T4 GPU",
        }

    @classmethod
    def return_type(cls):
        return str

    task: Task = Field(description="Choose a task.", default="image_captioning")
    image: ImageRef = Field(default=ImageRef(), description="Input image")
    caption: str | None = Field(
        title="Caption",
        description="Type caption for the input image for image text matching task.",
        default=None,
    )
    question: str | None = Field(
        title="Question",
        description="Type question for the input image for visual question answering task.",
        default=None,
    )


class Blip2(ReplicateNode):
    """Answers questions about images"""

    @classmethod
    def replicate_model_id(cls):
        return "andreasjansson/blip-2:f677695e5e89f8b236e52ecd1d3f01beb44c34606419bcc19345e046d8f786f9"

    @classmethod
    def get_hardware(cls):
        return "Nvidia A100 (40GB) GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://replicate.delivery/pbxt/IJEPmgAlL2zNBNDoRRKFegTEcxnlRhoQxlNjPHSZEy0pSIKn/gg_bridge.jpeg",
            "created_at": "2023-02-13T07:06:23.521189Z",
            "description": "Answers questions about images",
            "github_url": "https://github.com/daanelson/cog-blip-2",
            "license_url": None,
            "name": "blip-2",
            "owner": "andreasjansson",
            "paper_url": "https://arxiv.org/abs/2301.12597",
            "run_count": 20879500,
            "url": "https://replicate.com/andreasjansson/blip-2",
            "visibility": "public",
            "hardware": "Nvidia A100 (40GB) GPU",
        }

    @classmethod
    def return_type(cls):
        return str

    image: ImageRef = Field(
        default=ImageRef(), description="Input image to query or caption"
    )
    caption: bool = Field(
        title="Caption",
        description="Select if you want to generate image captions instead of asking questions",
        default=False,
    )
    context: str | None = Field(
        title="Context",
        description="Optional - previous questions and answers to be used as context for answering current question",
        default=None,
    )
    question: str = Field(
        title="Question",
        description="Question to ask about this image. Leave blank for captioning",
        default="What is this a picture of?",
    )
    temperature: float = Field(
        title="Temperature",
        description="Temperature for use with nucleus sampling",
        ge=0.5,
        le=1.0,
        default=1,
    )
    use_nucleus_sampling: bool = Field(
        title="Use Nucleus Sampling",
        description="Toggles the model using nucleus sampling to generate responses",
        default=False,
    )


class Clip_model_name(str, Enum):
    VIT_L_14_OPENAI = "ViT-L-14/openai"
    VIT_H_14_LAION2B_S32B_B79K = "ViT-H-14/laion2b_s32b_b79k"
    VIT_BIGG_14_LAION2B_S39B_B160K = "ViT-bigG-14/laion2b_s39b_b160k"


class ClipInterrogator(ReplicateNode):
    """The CLIP Interrogator is a prompt engineering tool that combines OpenAI's CLIP and Salesforce's BLIP to optimize text prompts to match a given image. Use the resulting prompts with text-to-image models like Stable Diffusion to create cool art!"""

    @classmethod
    def replicate_model_id(cls):
        return "pharmapsychotic/clip-interrogator:8151e1c9f47e696fa316146a2e35812ccf79cfc9eba05b11c7f450155102af70"

    @classmethod
    def get_hardware(cls):
        return "Nvidia T4 GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://replicate.delivery/pbxt/HrXsgowfhbZi3dImGZoIcvnz7oZfMtFY4UAEU8vBIakTd8JQ/watercolour-4799014_960_720.jpg",
            "created_at": "2022-10-28T17:47:38.473429Z",
            "description": "The CLIP Interrogator is a prompt engineering tool that combines OpenAI's CLIP and Salesforce's BLIP to optimize text prompts to match a given image. Use the resulting prompts with text-to-image models like Stable Diffusion to create cool art!",
            "github_url": "https://github.com/pharmapsychotic/clip-interrogator",
            "license_url": "https://github.com/pharmapsychotic/clip-interrogator/blob/main/LICENSE",
            "name": "clip-interrogator",
            "owner": "pharmapsychotic",
            "paper_url": None,
            "run_count": 1569987,
            "url": "https://replicate.com/pharmapsychotic/clip-interrogator",
            "visibility": "public",
            "hardware": "Nvidia T4 GPU",
        }

    @classmethod
    def return_type(cls):
        return str

    mode: Mode = Field(
        description="Prompt mode (best takes 10-20 seconds, fast takes 1-2 seconds).",
        default="best",
    )
    image: ImageRef = Field(default=ImageRef(), description="Input image")
    clip_model_name: Clip_model_name = Field(
        description="Choose ViT-L for Stable Diffusion 1, ViT-H for Stable Diffusion 2, or ViT-bigG for Stable Diffusion XL.",
        default="ViT-L-14/openai",
    )


class Llava13b(ReplicateNode):
    """Visual instruction tuning towards large language and vision models with GPT-4 level capabilities"""

    @classmethod
    def replicate_model_id(cls):
        return "yorickvp/llava-13b:b5f6212d032508382d61ff00469ddda3e32fd8a0e75dc39d8a4191bb742157fb"

    @classmethod
    def get_hardware(cls):
        return "Nvidia A40 (Large) GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://tjzk.replicate.delivery/models_models_cover_image/2c5dbfff-209d-4ab5-a294-b3e5e56105c0/dalle3.jpg",
            "created_at": "2023-10-09T16:27:51.777748Z",
            "description": "Visual instruction tuning towards large language and vision models with GPT-4 level capabilities",
            "github_url": "https://github.com/haotian-liu/LLaVA",
            "license_url": "https://ai.meta.com/llama/license/",
            "name": "llava-13b",
            "owner": "yorickvp",
            "paper_url": "https://arxiv.org/abs/2310.03744",
            "run_count": 6850285,
            "url": "https://replicate.com/yorickvp/llava-13b",
            "visibility": "public",
            "hardware": "Nvidia A40 (Large) GPU",
        }

    @classmethod
    def return_type(cls):
        return str

    image: ImageRef = Field(default=ImageRef(), description="Input image")
    top_p: float = Field(
        title="Top P",
        description="When decoding text, samples from the top p percentage of most likely tokens; lower to ignore less likely tokens",
        ge=0.0,
        le=1.0,
        default=1,
    )
    prompt: str | None = Field(
        title="Prompt", description="Prompt to use for text generation", default=None
    )
    max_tokens: int = Field(
        title="Max Tokens",
        description="Maximum number of tokens to generate. A word is generally 2-3 tokens",
        ge=0.0,
        default=1024,
    )
    temperature: float = Field(
        title="Temperature",
        description="Adjusts randomness of outputs, greater than 1 is random and 0 is deterministic",
        ge=0.0,
        default=0.2,
    )



import boto3
from typing import Optional
import logging
from botocore.client import Config


def _patch_headers(request, **kwargs):
    # https://github.com/boto/boto/issues/840
    request.headers.pop("Expect", None)


class AWSClient:
    """
    A client for interacting with AWS services.

    Main concern of the client is session and token management.
    """

    region_name: str | None
    endpoint_url: str | None
    access_key_id: str | None
    secret_access_key: str | None
    session_token: str | None
    log: logging.Logger | None

    def __init__(
        self,
        region_name: str | None,
        endpoint_url: str | None = None,
        access_key_id: str | None = None,
        secret_access_key: str | None = None,
        session_token: str | None = None,
        log: logging.Logger | None = None,
    ):
        self.region_name = region_name
        self.endpoint_url = endpoint_url
        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key
        self.session_token = session_token
        if log is None:
            log = logging.getLogger(__name__)
            log.setLevel(logging.INFO)
            log.addHandler(logging.StreamHandler())
        self.log = log

    def get_role(self):
        """
        Get the current IAM role.
        """
        return self.client("sts").get_caller_identity()["Arn"]

    def assume_role(
        self,
        role_arn: str,
        session_name="AssumeRoleSession1",
    ):
        """
        Assume a role in AWS. This is useful for running code locally, but
        accessing AWS resources as if you were in AWS.
        """
        credentials = self.client("sts").assume_role(
            RoleArn=role_arn, RoleSessionName=session_name, DurationSeconds=3600 * 8
        )["Credentials"]

        print("Using these credentials going forward:", credentials)

        self.access_key_id = credentials["AccessKeyId"]
        self.secret_access_key = credentials["SecretAccessKey"]
        self.session_token = credentials["SessionToken"]

    def session(self):
        """
        Return a boto3 session.

        If we have temporary credentials, use them. Otherwise,
        rely on boto3 to find credentials in the environment.
        """
        return boto3.session.Session(  # type: ignore
            region_name=self.region_name,
            aws_access_key_id=self.access_key_id,
            aws_secret_access_key=self.secret_access_key,
            aws_session_token=self.session_token,
        )

    def client(self, service_name: str, **kwargs):
        """
        Creates a boto3 client for the given service.

        Args:
            service_name: The name of the service.
            kwargs: Additional arguments to pass to the client.
        """
        client = self.session().client(
            service_name, endpoint_url=self.endpoint_url, **kwargs  # type: ignore
        )
        client.meta.events.register("before-send.s3.*", _patch_headers)
        return client

    def get_s3_storage(self, bucket: str, domain: Optional[str] = None):
        """
        Create an S3 storage for the given bucket.

        Args:
            bucket: The name of the bucket.
        """
        from nodetool.storage.s3_storage import S3Storage

        assert self.log is not None

        return S3Storage(
            bucket_name=bucket,
            client=self.client("s3", config=Config(signature_version="s3v4")),
            log=self.log,
            endpoint_url=self.endpoint_url,
            domain=domain,
        )



from ..utils import common_annotator_call, create_node_input_types
import comfy.model_management as model_management

class BAE_Normal_Map_Preprocessor:
    @classmethod
    def INPUT_TYPES(s):
        return create_node_input_types()

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "execute"

    CATEGORY = "ControlNet Preprocessors/Normal and Depth Estimators"

    def execute(self, image, resolution=512, **kwargs):
        from controlnet_aux.normalbae import NormalBaeDetector

        model = NormalBaeDetector.from_pretrained().to(model_management.get_torch_device())
        out = common_annotator_call(model, image, resolution=resolution)
        del model
        return (out,)

NODE_CLASS_MAPPINGS = {
    "BAE-NormalMapPreprocessor": BAE_Normal_Map_Preprocessor
}
NODE_DISPLAY_NAME_MAPPINGS = {
    "BAE-NormalMapPreprocessor": "BAE Normal Map"
}


# This is an improved version and model of HED edge detection with Apache License, Version 2.0.
# Please use this implementation in your products
# This implementation may produce slightly different results from Saining Xie's official implementations,
# but it generates smoother edges and is more suitable for ControlNet as well as other image-to-image translations.
# Different from official models and other implementations, this is an RGB-input model (rather than BGR)
# and in this way it works better for gradio's RGB protocol

import os
import warnings

import cv2
import numpy as np
import torch
from einops import rearrange
from PIL import Image

from controlnet_aux.util import HWC3, nms, resize_image_with_pad, safe_step, common_input_validate, custom_hf_download, HF_MODEL_NAME


class DoubleConvBlock(torch.nn.Module):
    def __init__(self, input_channel, output_channel, layer_number):
        super().__init__()
        self.convs = torch.nn.Sequential()
        self.convs.append(torch.nn.Conv2d(in_channels=input_channel, out_channels=output_channel, kernel_size=(3, 3), stride=(1, 1), padding=1))
        for i in range(1, layer_number):
            self.convs.append(torch.nn.Conv2d(in_channels=output_channel, out_channels=output_channel, kernel_size=(3, 3), stride=(1, 1), padding=1))
        self.projection = torch.nn.Conv2d(in_channels=output_channel, out_channels=1, kernel_size=(1, 1), stride=(1, 1), padding=0)

    def __call__(self, x, down_sampling=False):
        h = x
        if down_sampling:
            h = torch.nn.functional.max_pool2d(h, kernel_size=(2, 2), stride=(2, 2))
        for conv in self.convs:
            h = conv(h)
            h = torch.nn.functional.relu(h)
        return h, self.projection(h)


class ControlNetHED_Apache2(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.norm = torch.nn.Parameter(torch.zeros(size=(1, 3, 1, 1)))
        self.block1 = DoubleConvBlock(input_channel=3, output_channel=64, layer_number=2)
        self.block2 = DoubleConvBlock(input_channel=64, output_channel=128, layer_number=2)
        self.block3 = DoubleConvBlock(input_channel=128, output_channel=256, layer_number=3)
        self.block4 = DoubleConvBlock(input_channel=256, output_channel=512, layer_number=3)
        self.block5 = DoubleConvBlock(input_channel=512, output_channel=512, layer_number=3)

    def __call__(self, x):
        h = x - self.norm
        h, projection1 = self.block1(h)
        h, projection2 = self.block2(h, down_sampling=True)
        h, projection3 = self.block3(h, down_sampling=True)
        h, projection4 = self.block4(h, down_sampling=True)
        h, projection5 = self.block5(h, down_sampling=True)
        return projection1, projection2, projection3, projection4, projection5

class HEDdetector:
    def __init__(self, netNetwork):
        self.netNetwork = netNetwork

    @classmethod
    def from_pretrained(cls, pretrained_model_or_path=HF_MODEL_NAME, filename="ControlNetHED.pth"):
        model_path = custom_hf_download(pretrained_model_or_path, filename)

        netNetwork = ControlNetHED_Apache2()
        netNetwork.load_state_dict(torch.load(model_path, map_location='cpu'))
        netNetwork.float().eval()

        return cls(netNetwork)
    
    def to(self, device):
        self.netNetwork.to(device)
        return self
    

    def __call__(self, input_image, detect_resolution=512, safe=False, output_type="pil", scribble=False, upscale_method="INTER_CUBIC", **kwargs):
        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)
        input_image, remove_pad = resize_image_with_pad(input_image, detect_resolution, upscale_method)

        assert input_image.ndim == 3
        H, W, C = input_image.shape
        with torch.no_grad():
            device = next(iter(self.netNetwork.parameters())).device
            image_hed = torch.from_numpy(input_image).float().to(device)
            image_hed = rearrange(image_hed, 'h w c -> 1 c h w')
            edges = self.netNetwork(image_hed)
            edges = [e.detach().cpu().numpy().astype(np.float32)[0, 0] for e in edges]
            edges = [cv2.resize(e, (W, H), interpolation=cv2.INTER_LINEAR) for e in edges]
            edges = np.stack(edges, axis=2)
            edge = 1 / (1 + np.exp(-np.mean(edges, axis=2).astype(np.float64)))
            if safe:
                edge = safe_step(edge)
            edge = (edge * 255.0).clip(0, 255).astype(np.uint8)

        detected_map = edge
        
        if scribble:
            detected_map = nms(detected_map, 127, 3.0)
            detected_map = cv2.GaussianBlur(detected_map, (0, 0), 3.0)
            detected_map[detected_map > 4] = 255
            detected_map[detected_map < 255] = 0

        detected_map = HWC3(remove_pad(detected_map))

        if output_type == "pil":
            detected_map = Image.fromarray(detected_map)

        return detected_map



import torch


def compute_rotation_matrix_from_ortho6d(poses):
    """
    Code from
    https://github.com/papagina/RotationContinuity
    On the Continuity of Rotation Representations in Neural Networks
    Zhou et al. CVPR19
    https://zhouyisjtu.github.io/project_rotation/rotation.html
    """
    x_raw = poses[:, 0:3]  # batch*3
    y_raw = poses[:, 3:6]  # batch*3
        
    x = normalize_vector(x_raw)  # batch*3
    z = cross_product(x, y_raw)  # batch*3
    z = normalize_vector(z)  # batch*3
    y = cross_product(z, x)  # batch*3
        
    x = x.view(-1, 3, 1)
    y = y.view(-1, 3, 1)
    z = z.view(-1, 3, 1)
    matrix = torch.cat((x, y, z), 2)  # batch*3*3
    return matrix

def robust_compute_rotation_matrix_from_ortho6d(poses):
    """
    Instead of making 2nd vector orthogonal to first
    create a base that takes into account the two predicted
    directions equally
    """
    x_raw = poses[:, 0:3]  # batch*3
    y_raw = poses[:, 3:6]  # batch*3

    x = normalize_vector(x_raw)  # batch*3
    y = normalize_vector(y_raw)  # batch*3
    middle = normalize_vector(x + y)
    orthmid = normalize_vector(x - y)
    x = normalize_vector(middle + orthmid)
    y = normalize_vector(middle - orthmid)
    # Their scalar product should be small !
    # assert torch.einsum("ij,ij->i", [x, y]).abs().max() < 0.00001
    z = normalize_vector(cross_product(x, y))

    x = x.view(-1, 3, 1)
    y = y.view(-1, 3, 1)
    z = z.view(-1, 3, 1)
    matrix = torch.cat((x, y, z), 2)  # batch*3*3
    # Check for reflection in matrix ! If found, flip last vector TODO
    assert (torch.stack([torch.det(mat) for mat in matrix ])< 0).sum() == 0
    return matrix


def normalize_vector(v):
    batch = v.shape[0]
    v_mag = torch.sqrt(v.pow(2).sum(1))  # batch
    v_mag = torch.max(v_mag, v.new([1e-8]))
    v_mag = v_mag.view(batch, 1).expand(batch, v.shape[1])
    v = v/v_mag
    return v


def cross_product(u, v):
    batch = u.shape[0]
    i = u[:, 1] * v[:, 2] - u[:, 2] * v[:, 1]
    j = u[:, 2] * v[:, 0] - u[:, 0] * v[:, 2]
    k = u[:, 0] * v[:, 1] - u[:, 1] * v[:, 0]
        
    out = torch.cat((i.view(batch, 1), j.view(batch, 1), k.view(batch, 1)), 1)
        
    return out



import torch
from comfy.model_management import get_torch_device

device = get_torch_device()
#https://github.com/microsoft/DirectML/issues/414#issuecomment-1541319479
def sparse_to_dense(sparse_tensor):
    if device.type in ['dml', 'mps', 'xpu']:
        return sparse_tensor.to_dense()
    return sparse_tensor


#!/usr/bin/env python

from datetime import datetime
from fastapi import APIRouter, Depends, HTTPException, Body
from nodetool.api.utils import current_user, User
from nodetool.api.types.chat import MessageCreateRequest, MessageList
from nodetool.metadata.types import Message
from nodetool.models.message import Message as MessageModel
from nodetool.common.environment import Environment
from typing import Optional

from nodetool.models.thread import Thread


log = Environment.get_logger()
router = APIRouter(prefix="/api/messages", tags=["messages"])


@router.post("/")
async def create(
    req: MessageCreateRequest, user: User = Depends(current_user)
) -> Message:
    if req.thread_id is None:
        thread_id = Thread.create(user_id=user.id).id
    else:
        thread_id = req.thread_id
    return Message.from_model(
        MessageModel.create(
            user_id=user.id,
            thread_id=thread_id,
            tool_call_id=req.tool_call_id,
            role=req.role,
            name=req.name,
            content=req.content,
            tool_calls=req.tool_calls,
            created_at=datetime.now(),
        )
    )


@router.get("/{message_id}")
async def get(message_id: str, user: User = Depends(current_user)) -> Message:
    message = MessageModel.get(message_id)
    if message is None:
        raise HTTPException(status_code=404, detail="Message not found")
    if message.user_id != user.id:
        raise HTTPException(status_code=404, detail="Message not found")
    return Message.from_model(message)


@router.get("/")
async def index(
    thread_id: str,
    reverse: bool = False,
    user: User = Depends(current_user),
    cursor: Optional[str] = None,
    limit: int = 100,
) -> MessageList:
    messages, cursor = MessageModel.paginate(
        thread_id=thread_id, reverse=reverse, limit=limit, start_key=cursor
    )
    for message in messages:
        if message.user_id != user.id:
            raise HTTPException(status_code=404, detail="Message not found")

    return MessageList(
        next=cursor, messages=[Message.from_model(message) for message in messages]
    )



from __future__ import division
import torch
import torch.nn.functional as F
import numpy as np
import scipy.sparse
import math
from pathlib import Path
data_path = Path(__file__).parent / "data"

from comfy.model_management import get_torch_device
from wrapper_for_mps import sparse_to_dense
device = get_torch_device()

class SparseMM(torch.autograd.Function):
    """Redefine sparse @ dense matrix multiplication to enable backpropagation.
    The builtin matrix multiplication operation does not support backpropagation in some cases.
    """
    @staticmethod
    def forward(ctx, sparse, dense):
        ctx.req_grad = dense.requires_grad
        ctx.save_for_backward(sparse)
        return torch.matmul(sparse, dense)

    @staticmethod
    def backward(ctx, grad_output):
        grad_input = None
        sparse, = ctx.saved_tensors
        if ctx.req_grad:
            grad_input = torch.matmul(sparse.t(), grad_output)
        return None, grad_input

def spmm(sparse, dense):
    sparse = sparse.to(device)
    dense = dense.to(device)
    return SparseMM.apply(sparse, dense)


def gelu(x):
    """Implementation of the gelu activation function.
        For information: OpenAI GPT's gelu is slightly different (and gives slightly different results):
        0.5 * x * (1 + torch.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3))))
        Also see https://arxiv.org/abs/1606.08415
    """
    return x * 0.5 * (1.0 + torch.erf(x / math.sqrt(2.0)))

class BertLayerNorm(torch.nn.Module):
    def __init__(self, hidden_size, eps=1e-12):
        """Construct a layernorm module in the TF style (epsilon inside the square root).
        """
        super(BertLayerNorm, self).__init__()
        self.weight = torch.nn.Parameter(torch.ones(hidden_size))
        self.bias = torch.nn.Parameter(torch.zeros(hidden_size))
        self.variance_epsilon = eps

    def forward(self, x):
        u = x.mean(-1, keepdim=True)
        s = (x - u).pow(2).mean(-1, keepdim=True)
        x = (x - u) / torch.sqrt(s + self.variance_epsilon)
        return self.weight * x + self.bias


class GraphResBlock(torch.nn.Module):
    """
    Graph Residual Block similar to the Bottleneck Residual Block in ResNet
    """
    def __init__(self, in_channels, out_channels, mesh_type='body'):
        super(GraphResBlock, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.lin1 = GraphLinear(in_channels, out_channels // 2)
        self.conv = GraphConvolution(out_channels // 2, out_channels // 2, mesh_type)
        self.lin2 = GraphLinear(out_channels // 2, out_channels)
        self.skip_conv = GraphLinear(in_channels, out_channels)
        # print('Use BertLayerNorm in GraphResBlock')
        self.pre_norm = BertLayerNorm(in_channels)
        self.norm1 = BertLayerNorm(out_channels // 2)
        self.norm2 = BertLayerNorm(out_channels // 2)

    def forward(self, x):
        trans_y = F.relu(self.pre_norm(x)).transpose(1,2)
        y = self.lin1(trans_y).transpose(1,2)

        y = F.relu(self.norm1(y))
        y = self.conv(y)

        trans_y = F.relu(self.norm2(y)).transpose(1,2)
        y = self.lin2(trans_y).transpose(1,2)

        z = x+y

        return z

# class GraphResBlock(torch.nn.Module):
#     """
#     Graph Residual Block similar to the Bottleneck Residual Block in ResNet
#     """
#     def __init__(self, in_channels, out_channels, mesh_type='body'):
#         super(GraphResBlock, self).__init__()
#         self.in_channels = in_channels
#         self.out_channels = out_channels
#         self.conv = GraphConvolution(self.in_channels, self.out_channels, mesh_type)
#         print('Use BertLayerNorm and GeLU in GraphResBlock')
#         self.norm = BertLayerNorm(self.out_channels)
#     def forward(self, x):
#         y = self.conv(x)
#         y = self.norm(y)
#         y = gelu(y)
#         z = x+y
#         return z

class GraphLinear(torch.nn.Module):
    """
    Generalization of 1x1 convolutions on Graphs
    """
    def __init__(self, in_channels, out_channels):
        super(GraphLinear, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.W = torch.nn.Parameter(torch.FloatTensor(out_channels, in_channels))
        self.b = torch.nn.Parameter(torch.FloatTensor(out_channels))
        self.reset_parameters()

    def reset_parameters(self):
        w_stdv = 1 / (self.in_channels * self.out_channels)
        self.W.data.uniform_(-w_stdv, w_stdv)
        self.b.data.uniform_(-w_stdv, w_stdv)

    def forward(self, x):
        return torch.matmul(self.W[None, :], x) + self.b[None, :, None]

class GraphConvolution(torch.nn.Module):
    """Simple GCN layer, similar to https://arxiv.org/abs/1609.02907."""
    def __init__(self, in_features, out_features, mesh='body', bias=True):
        super(GraphConvolution, self).__init__()
        self.in_features = in_features
        self.out_features = out_features

        if mesh=='body':
            adj_indices = torch.load(data_path / 'smpl_431_adjmat_indices.pt')
            adj_mat_value = torch.load(data_path / 'smpl_431_adjmat_values.pt')
            adj_mat_size = torch.load(data_path / 'smpl_431_adjmat_size.pt')
        elif mesh=='hand':
            adj_indices = torch.load(data_path / 'mano_195_adjmat_indices.pt')
            adj_mat_value = torch.load(data_path / 'mano_195_adjmat_values.pt')
            adj_mat_size = torch.load(data_path / 'mano_195_adjmat_size.pt')

        self.adjmat = sparse_to_dense(torch.sparse_coo_tensor(adj_indices, adj_mat_value, size=adj_mat_size)).to(device)

        self.weight = torch.nn.Parameter(torch.FloatTensor(in_features, out_features))
        if bias:
            self.bias = torch.nn.Parameter(torch.FloatTensor(out_features))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self):
        # stdv = 1. / math.sqrt(self.weight.size(1))
        stdv = 6. / math.sqrt(self.weight.size(0) + self.weight.size(1))
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)

    def forward(self, x):
        if x.ndimension() == 2:
            support = torch.matmul(x, self.weight)
            output = torch.matmul(self.adjmat, support)
            if self.bias is not None:
                output = output + self.bias
            return output
        else:
            output = []
            for i in range(x.shape[0]):
                support = torch.matmul(x[i], self.weight)
                # output.append(torch.matmul(self.adjmat, support))
                output.append(spmm(self.adjmat, support))
            output = torch.stack(output, dim=0)
            if self.bias is not None:
                output = output + self.bias
            return output

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + str(self.in_features) + ' -> ' \
               + str(self.out_features) + ')'


from pydantic import Field
from nodetool.metadata.types import AudioRef
from nodetool.metadata.types import ImageRef
from nodetool.common.huggingface_node import HuggingfaceNode
from nodetool.workflows.processing_context import ProcessingContext
from enum import Enum


class ModelId(str, Enum):
    FASTSPEECH2_EN_LJSPEECH = "facebook/fastspeech2-en-ljspeech"
    SUNO_BARK = "suno/bark"


class TextToSpeech(HuggingfaceNode):
    """
    Text-to-Speech (TTS) is the task of generating natural sounding
    speech given text input. TTS models can be extended to have a single model
    that generates speech for multiple speakers and multiple languages.

    ### Use Cases
    * Text-to-Speech (TTS) models can be used in any speech-enabled application that requires converting text to speech imitating human voice.
    * TTS models are used to create voice assistants on smart devices. These models are a better alternative compared to concatenative methods where the assistant is built by recording sounds and mapping them, since the outputs in TTS models contain elements in natural speech such as emphasis.
    * TTS models are widely used in airport and public transportation announcement systems to convert the announcement of a given text into speech.

    """

    model: ModelId = Field(
        default=ModelId.SUNO_BARK,
        title="Model ID on Huggingface",
        description="The model ID to use for the image generation",
    )
    inputs: str = Field(
        title="Inputs",
        description="The input text to the model",
    )

    async def process(self, context: ProcessingContext) -> AudioRef:
        result = await self.run_huggingface(
            model_id=self.model.value, context=context, params={"inputs": self.inputs}
        )
        audio = await context.audio_from_bytes(result)  # type: ignore
        return audio


class TextToAudioModel(str, Enum):
    MUSICGEN_SMALL = "facebook/musicgen-small"
    MUSICGEN_MEDIUM = "facebook/musicgen-medium"
    MUSICGEN_LARGE = "facebook/musicgen-large"
    MUSICGEN_MELODY = "facebook/musicgen-melody"
    MUSICGEN_STEREO_SMALL = "facebook/musicgen-stereo-small"
    MUSICGEN_STEREO_LARGE = "facebook/musicgen-stereo-large"


class TextToAudio(HuggingfaceNode):
    """
    Generate audio from a text input using a Huggingface model.
    audio, text, huggingface
    """

    model: TextToAudioModel = Field(
        default=TextToAudioModel.MUSICGEN_SMALL,
        title="Model ID on Huggingface",
        description="The model ID to use for the audio generation",
    )
    inputs: str = Field(
        default="",
        title="Inputs",
        description="The input text to the model",
    )

    async def process(self, context: ProcessingContext) -> AudioRef:
        result = await self.run_huggingface(
            model_id=self.model.value,
            context=context,
            params={
                "inputs": self.inputs,
            },
        )
        audio = await context.audio_from_bytes(result)  # type: ignore
        return audio



# Source: https://github.com/Hyuto/yolo-nas-onnx/tree/master/yolo-nas-py
# Inspired from: https://github.com/Deci-AI/super-gradients/blob/3.1.1/src/super_gradients/training/processing/processing.py

import numpy as np
import cv2

def preprocess(img, input_size, swap=(2, 0, 1)):
    if len(img.shape) == 3:
        padded_img = np.ones((input_size[0], input_size[1], 3), dtype=np.uint8) * 114
    else:
        padded_img = np.ones(input_size, dtype=np.uint8) * 114

    r = min(input_size[0] / img.shape[0], input_size[1] / img.shape[1])
    resized_img = cv2.resize(
        img,
        (int(img.shape[1] * r), int(img.shape[0] * r)),
        interpolation=cv2.INTER_LINEAR,
    ).astype(np.uint8)
    padded_img[: int(img.shape[0] * r), : int(img.shape[1] * r)] = resized_img

    padded_img = padded_img.transpose(swap)
    padded_img = np.ascontiguousarray(padded_img, dtype=np.float32)
    return padded_img, r

def inference_detector(session, oriImg, detect_classes=[0], dtype=np.uint8):
    """
    This function is only compatible with onnx models exported from the new API with built-in NMS
    ```py
    from super_gradients.conversion.conversion_enums import ExportQuantizationMode
    from super_gradients.common.object_names import Models
    from super_gradients.training import models

    model = models.get(Models.YOLO_NAS_L, pretrained_weights="coco")

    export_result = model.export(
        "yolo_nas/yolo_nas_l_fp16.onnx",
        quantization_mode=ExportQuantizationMode.FP16,
        device="cuda"
    )
    ```
    """
    input_shape = (640,640)
    img, ratio = preprocess(oriImg, input_shape)
    input = img[None, :, :, :]
    input = input.astype(dtype)
    if "InferenceSession" in type(session).__name__:
        input_name = session.get_inputs()[0].name
        output = session.run(None, {input_name: input})
    else:
        outNames = session.getUnconnectedOutLayersNames()
        session.setInput(input)
        output = session.forward(outNames)
    num_preds, pred_boxes, pred_scores, pred_classes = output
    num_preds = num_preds[0,0]
    if num_preds == 0:
        return None
    idxs = np.where((np.isin(pred_classes[0, :num_preds], detect_classes)) & (pred_scores[0, :num_preds] > 0.3))
    if (len(idxs) == 0) or (idxs[0].size == 0):
        return None
    return pred_boxes[0, idxs].squeeze(axis=0) / ratio



from typing import Literal, Optional, List
from pydantic import BaseModel
from nodetool.models.job import Job
from pydantic import BaseModel, Field


class JobRequest(BaseModel):
    workflow_id: str
    params: dict


class JobUpdate(BaseModel):
    type: Literal["job_update"] = "job_update"
    status: str
    error: str | None = None


class JobList(BaseModel):
    next: Optional[str]
    jobs: List[Job]



# ------------------------------------------------------------------------------
# Copyright (c) Microsoft
# Licensed under the MIT License.
# Create by Bin Xiao (Bin.Xiao@microsoft.com)
# Modified by Ke Sun (sunk@mail.ustc.edu.cn)
# ------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from yacs.config import CfgNode as CN

# high_resoluton_net related params for classification
POSE_HIGH_RESOLUTION_NET = CN()
POSE_HIGH_RESOLUTION_NET.PRETRAINED_LAYERS = ['*']
POSE_HIGH_RESOLUTION_NET.STEM_INPLANES = 64
POSE_HIGH_RESOLUTION_NET.FINAL_CONV_KERNEL = 1
POSE_HIGH_RESOLUTION_NET.WITH_HEAD = True

POSE_HIGH_RESOLUTION_NET.STAGE2 = CN()
POSE_HIGH_RESOLUTION_NET.STAGE2.NUM_MODULES = 1
POSE_HIGH_RESOLUTION_NET.STAGE2.NUM_BRANCHES = 2
POSE_HIGH_RESOLUTION_NET.STAGE2.NUM_BLOCKS = [4, 4]
POSE_HIGH_RESOLUTION_NET.STAGE2.NUM_CHANNELS = [32, 64]
POSE_HIGH_RESOLUTION_NET.STAGE2.BLOCK = 'BASIC'
POSE_HIGH_RESOLUTION_NET.STAGE2.FUSE_METHOD = 'SUM'

POSE_HIGH_RESOLUTION_NET.STAGE3 = CN()
POSE_HIGH_RESOLUTION_NET.STAGE3.NUM_MODULES = 1
POSE_HIGH_RESOLUTION_NET.STAGE3.NUM_BRANCHES = 3
POSE_HIGH_RESOLUTION_NET.STAGE3.NUM_BLOCKS = [4, 4, 4]
POSE_HIGH_RESOLUTION_NET.STAGE3.NUM_CHANNELS = [32, 64, 128]
POSE_HIGH_RESOLUTION_NET.STAGE3.BLOCK = 'BASIC'
POSE_HIGH_RESOLUTION_NET.STAGE3.FUSE_METHOD = 'SUM'

POSE_HIGH_RESOLUTION_NET.STAGE4 = CN()
POSE_HIGH_RESOLUTION_NET.STAGE4.NUM_MODULES = 1
POSE_HIGH_RESOLUTION_NET.STAGE4.NUM_BRANCHES = 4
POSE_HIGH_RESOLUTION_NET.STAGE4.NUM_BLOCKS = [4, 4, 4, 4]
POSE_HIGH_RESOLUTION_NET.STAGE4.NUM_CHANNELS = [32, 64, 128, 256]
POSE_HIGH_RESOLUTION_NET.STAGE4.BLOCK = 'BASIC'
POSE_HIGH_RESOLUTION_NET.STAGE4.FUSE_METHOD = 'SUM'

MODEL_EXTRAS = {
    'cls_hrnet': POSE_HIGH_RESOLUTION_NET,
}



from pydantic import BaseModel, Field
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode
from nodetool.common.replicate_node import ReplicateNode
from enum import Enum


class AudioSuperResolution(ReplicateNode):
    """AudioSR: Versatile Audio Super-resolution at Scale"""

    @classmethod
    def replicate_model_id(cls):
        return "nateraw/audio-super-resolution:9c3d3e39fb0cb6aea677264881d8073f835336137b39fdea4e94093319379535"

    @classmethod
    def get_hardware(cls):
        return "Nvidia A40 (Large) GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://tjzk.replicate.delivery/models_models_cover_image/6bc4e480-6695-451b-940d-48a3b83a1356/replicate-prediction-jvi5xvlbg4v4.png",
            "created_at": "2023-09-20T04:53:52.943393Z",
            "description": "AudioSR: Versatile Audio Super-resolution at Scale",
            "github_url": "https://github.com/haoheliu/versatile_audio_super_resolution",
            "license_url": "https://huggingface.co/haoheliu/wellsolve_audio_super_resolution_48k/blob/main/README.md",
            "name": "audio-super-resolution",
            "owner": "nateraw",
            "paper_url": "https://arxiv.org/abs/2309.07314",
            "run_count": 36587,
            "url": "https://replicate.com/nateraw/audio-super-resolution",
            "visibility": "public",
            "hardware": "Nvidia A40 (Large) GPU",
        }

    @classmethod
    def return_type(cls):
        return AudioRef

    seed: int | None = Field(
        title="Seed",
        description="Random seed. Leave blank to randomize the seed",
        default=None,
    )
    ddim_steps: int = Field(
        title="Ddim Steps",
        description="Number of inference steps",
        ge=10.0,
        le=500.0,
        default=50,
    )
    input_file: AudioRef = Field(default=AudioRef(), description="Audio to upsample")
    guidance_scale: float = Field(
        title="Guidance Scale",
        description="Scale for classifier free guidance",
        ge=1.0,
        le=20.0,
        default=3.5,
    )



# ------------------------------------------------------------------------------
# Reference: https://github.com/facebookresearch/detectron2/blob/main/detectron2/evaluation/cityscapes_evaluation.py
# Modified by Jitesh Jain (https://github.com/praeclarumjj3)
# ------------------------------------------------------------------------------

import glob
import logging
import numpy as np
import os
import tempfile
from collections import OrderedDict
import torch
from PIL import Image

from custom_detectron2.data import MetadataCatalog
from custom_detectron2.utils import comm
from custom_detectron2.utils.file_io import PathManager

from .evaluator import DatasetEvaluator


class CityscapesEvaluator(DatasetEvaluator):
    """
    Base class for evaluation using cityscapes API.
    """

    def __init__(self, dataset_name):
        """
        Args:
            dataset_name (str): the name of the dataset.
                It must have the following metadata associated with it:
                "thing_classes", "gt_dir".
        """
        self._metadata = MetadataCatalog.get(dataset_name)
        self._cpu_device = torch.device("cpu")
        self._logger = logging.getLogger(__name__)

    def reset(self):
        self._working_dir = tempfile.TemporaryDirectory(prefix="cityscapes_eval_")
        self._temp_dir = self._working_dir.name
        # All workers will write to the same results directory
        # TODO this does not work in distributed training
        assert (
            comm.get_local_size() == comm.get_world_size()
        ), "CityscapesEvaluator currently do not work with multiple machines."
        self._temp_dir = comm.all_gather(self._temp_dir)[0]
        if self._temp_dir != self._working_dir.name:
            self._working_dir.cleanup()
        self._logger.info(
            "Writing cityscapes results to temporary directory {} ...".format(self._temp_dir)
        )


class CityscapesInstanceEvaluator(CityscapesEvaluator):
    """
    Evaluate instance segmentation results on cityscapes dataset using cityscapes API.

    Note:
        * It does not work in multi-machine distributed training.
        * It contains a synchronization, therefore has to be used on all ranks.
        * Only the main process runs evaluation.
    """

    def process(self, inputs, outputs):
        from cityscapesscripts.helpers.labels import name2label

        for input, output in zip(inputs, outputs):
            file_name = input["file_name"]
            basename = os.path.splitext(os.path.basename(file_name))[0]
            pred_txt = os.path.join(self._temp_dir, basename + "_pred.txt")

            if "instances" in output:
                output = output["instances"].to(self._cpu_device)
                num_instances = len(output)
                with open(pred_txt, "w") as fout:
                    for i in range(num_instances):
                        pred_class = output.pred_classes[i]
                        classes = self._metadata.stuff_classes[pred_class]
                        class_id = name2label[classes].id
                        score = output.scores[i]
                        mask = output.pred_masks[i].numpy().astype("uint8")
                        png_filename = os.path.join(
                            self._temp_dir, basename + "_{}_{}.png".format(i, classes)
                        )

                        Image.fromarray(mask * 255).save(png_filename)
                        fout.write(
                            "{} {} {}\n".format(os.path.basename(png_filename), class_id, score)
                        )
            else:
                # Cityscapes requires a prediction file for every ground truth image.
                with open(pred_txt, "w") as fout:
                    pass

    def evaluate(self):
        """
        Returns:
            dict: has a key "segm", whose value is a dict of "AP" and "AP50".
        """
        comm.synchronize()
        if comm.get_rank() > 0:
            return
        import cityscapesscripts.evaluation.evalInstanceLevelSemanticLabeling as cityscapes_eval

        self._logger.info("Evaluating results under {} ...".format(self._temp_dir))

        # set some global states in cityscapes evaluation API, before evaluating
        cityscapes_eval.args.predictionPath = os.path.abspath(self._temp_dir)
        cityscapes_eval.args.predictionWalk = None
        cityscapes_eval.args.JSONOutput = False
        cityscapes_eval.args.colorized = False
        cityscapes_eval.args.gtInstancesFile = os.path.join(self._temp_dir, "gtInstances.json")

        # These lines are adopted from
        # https://github.com/mcordts/cityscapesScripts/blob/master/cityscapesscripts/evaluation/evalInstanceLevelSemanticLabeling.py # noqa
        gt_dir = PathManager.get_local_path(self._metadata.gt_dir)
        groundTruthImgList = glob.glob(os.path.join(gt_dir, "*", "*_gtFine_instanceIds.png"))
        assert len(
            groundTruthImgList
        ), "Cannot find any ground truth images to use for evaluation. Searched for: {}".format(
            cityscapes_eval.args.groundTruthSearch
        )
        predictionImgList = []
        for gt in groundTruthImgList:
            predictionImgList.append(cityscapes_eval.getPrediction(gt, cityscapes_eval.args))
        results = cityscapes_eval.evaluateImgLists(
            predictionImgList, groundTruthImgList, cityscapes_eval.args
        )["averages"]

        ret = OrderedDict()
        ret["segm"] = {"AP": results["allAp"] * 100, "AP50": results["allAp50%"] * 100}
        self._working_dir.cleanup()
        return ret


class CityscapesSemSegEvaluator(CityscapesEvaluator):
    """
    Evaluate semantic segmentation results on cityscapes dataset using cityscapes API.

    Note:
        * It does not work in multi-machine distributed training.
        * It contains a synchronization, therefore has to be used on all ranks.
        * Only the main process runs evaluation.
    """

    def process(self, inputs, outputs):
        from cityscapesscripts.helpers.labels import trainId2label

        for input, output in zip(inputs, outputs):
            file_name = input["file_name"]
            basename = os.path.splitext(os.path.basename(file_name))[0]
            pred_filename = os.path.join(self._temp_dir, basename + "_pred.png")

            output = output["sem_seg"].argmax(dim=0).to(self._cpu_device).numpy()
            pred = 255 * np.ones(output.shape, dtype=np.uint8)
            for train_id, label in trainId2label.items():
                if label.ignoreInEval:
                    continue
                pred[output == train_id] = label.id
            Image.fromarray(pred).save(pred_filename)

    def evaluate(self):
        comm.synchronize()
        if comm.get_rank() > 0:
            return
        # Load the Cityscapes eval script *after* setting the required env var,
        # since the script reads CITYSCAPES_DATASET into global variables at load time.
        import cityscapesscripts.evaluation.evalPixelLevelSemanticLabeling as cityscapes_eval

        self._logger.info("Evaluating results under {} ...".format(self._temp_dir))

        # set some global states in cityscapes evaluation API, before evaluating
        cityscapes_eval.args.predictionPath = os.path.abspath(self._temp_dir)
        cityscapes_eval.args.predictionWalk = None
        cityscapes_eval.args.JSONOutput = False
        cityscapes_eval.args.colorized = False

        # These lines are adopted from
        # https://github.com/mcordts/cityscapesScripts/blob/master/cityscapesscripts/evaluation/evalPixelLevelSemanticLabeling.py # noqa
        gt_dir = PathManager.get_local_path(self._metadata.gt_dir)
        groundTruthImgList = glob.glob(os.path.join(gt_dir, "*", "*_gtFine_labelIds.png"))
        assert len(
            groundTruthImgList
        ), "Cannot find any ground truth images to use for evaluation. Searched for: {}".format(
            cityscapes_eval.args.groundTruthSearch
        )
        predictionImgList = []
        for gt in groundTruthImgList:
            predictionImgList.append(cityscapes_eval.getPrediction(cityscapes_eval.args, gt))
        results = cityscapes_eval.evaluateImgLists(
            predictionImgList, groundTruthImgList, cityscapes_eval.args
        )
        ret = OrderedDict()
        ret["sem_seg"] = {
            "IoU": 100.0 * results["averageScoreClasses"],
            "iIoU": 100.0 * results["averageScoreInstClasses"],
            "IoU_sup": 100.0 * results["averageScoreCategories"],
            "iIoU_sup": 100.0 * results["averageScoreInstCategories"],
        }
        self._working_dir.cleanup()
        return ret



import base64
from datetime import datetime
import enum
import json
from types import UnionType
from typing import Any, Dict, Type, Union, get_origin
from pydantic.fields import FieldInfo

from nodetool.common.environment import Environment

from .database_adapter import DatabaseAdapter

log = Environment.get_logger()


def is_union_type(t):
    """
    Check if a type is a union.

    Args:
        t: The type to check.

    Returns:
        True if the type is a union, False otherwise.
    """
    return get_origin(t) is Union or isinstance(t, UnionType)


def convert_to_dynamo(value: Any) -> dict | None:
    """
    Convert a Python value to its DynamoDB representation.

    :param value: The value to convert.
    :return: A dictionary representing the DynamoDB format of the value.
    """
    py_type = type(value)

    if value is None:
        raise ValueError("Cannot convert None to DynamoDB format")

    return convert_to_dynamodb_format(value, py_type)


def convert_to_dynamodb_format(
    value: Any,
    py_type: Type,
) -> dict | None:
    """
    Convert a Python value to its DynamoDB representation based on the provided Python type.
    Use default_value if the provided value is None. Convert lists and dicts to JSON strings.

    :param value: The value to convert, or None.
    :param py_type: The Python type of the value.
    """
    if value is None:
        return None

    # unwrap union and optional types
    if is_union_type(py_type):
        args = [t for t in py_type.__args__ if t != type(None)]
        if len(args) == 1:
            return convert_to_dynamodb_format(value, args[0])
        else:
            return {"S": json.dumps(value)}

    if py_type == Any:
        return convert_to_dynamo(value)
    elif py_type == bytes:
        return {"B": base64.b64encode(value).decode()}
    elif get_origin(py_type) == list:
        return {"L": [convert_to_dynamodb_format(v, py_type.__args__[0]) for v in value]}  # type: ignore
    elif get_origin(py_type) == dict:
        if py_type.__args__[0] != str:  # type: ignore
            raise TypeError(f"Unsupported key type for DynamoDB Map: {py_type.__args__[0]}")  # type: ignore
        return {"M": {k: convert_to_dynamodb_format(v, py_type.__args__[1]) for k, v in value.items()}}  # type: ignore
    elif get_origin(py_type) == set:
        if len(value) == 0:
            return None
        if py_type.__args__[0] in {int, float}:  # type: ignore
            return {"NS": [str(v) for v in value]}
        elif py_type.__args__[0] == bytes:  # type: ignore
            return {"BS": [base64.b64encode(v).decode() for v in value]}
        elif py_type.__args__[0] == str:  # type: ignore
            return {"SS": [str(v) for v in value]}
    elif py_type == list:
        return {"S": json.dumps(value)}
    elif py_type == dict:
        return {"S": json.dumps(value)}
    elif py_type == str:
        return {"S": value}
    elif py_type == int:
        return {"N": str(value)}
    elif py_type == float:
        return {"N": str(value)}
    elif py_type == bool:
        return {"BOOL": value is True}
    elif py_type == datetime:
        return {"S": value.isoformat()}
    elif issubclass(py_type, enum.Enum):
        return convert_to_dynamo(value.value)
    else:
        raise TypeError(f"Unsupported type for DynamoDB: {py_type}")


def convert_from_dynamodb_format(item: Dict[str, Any], py_type: Type) -> Any:
    """
    Convert a value from its DynamoDB representation to the specified Python type.

    :param item: A dictionary representing the DynamoDB format of the value.
    :param py_type: The Python type to convert to.
    :return: The Python value corresponding to the DynamoDB item.
    """
    # unwrap union and optional types
    if is_union_type(py_type):
        args = [t for t in py_type.__args__ if t != type(None)]
        if len(args) == 1:
            return convert_from_dynamodb_format(item, args[0])
        else:
            json_str = item.get("S", "{}")
            return json.loads(json_str) if json_str else {}

    if get_origin(py_type) == set and py_type.__args__[0] in {int, float}:  # type: ignore
        return {py_type.__args__[0](v) for v in item.get("NS", [])}  # type: ignore
    elif get_origin(py_type) == set and py_type.__args__[0] == bytes:  # type: ignore
        return {base64.b64decode(v) for v in item.get("BS", [])}
    elif get_origin(py_type) == set and py_type.__args__[0] == str:  # type: ignore
        return {str(v) for v in item.get("SS", [])}
    elif get_origin(py_type) == dict:
        if py_type.__args__[0] != str:  # type: ignore
            raise TypeError(
                f"Unsupported key type for DynamoDB Map: {py_type.__args__[0]}"  # type: ignore
            )
        return {k: convert_from_dynamodb_format(v, py_type.__args__[1]) for k, v in item.get("M", {}).items()}  # type: ignore
    elif get_origin(py_type) == list:
        return [
            convert_from_dynamodb_format(v, py_type.__args__[0])  # type: ignore
            for v in item.get("L", [])
        ]
    elif py_type == str:
        return item.get("S", None)
    elif py_type in [int, float]:
        return py_type(item.get("N", 0))
    elif py_type == bool:
        return item.get("BOOL", False)
    elif py_type == dict:
        json_str = item.get("S", "{}")
        return json.loads(json_str) if json_str else {}
    elif py_type == list:
        json_str = item.get("S", "[]")
        return json.loads(json_str) if json_str else []
    elif py_type == datetime:
        date_str = item.get("S", None)
        try:
            return datetime.fromisoformat(date_str) if date_str else None
        except ValueError:
            return datetime.strptime(date_str, "%Y-%m-%dT%H:%M:%S.%f%z")
    elif py_type == Any:
        json_str = item.get("S", "{}")
        try:
            return json.loads(json_str) if json_str else {}
        except json.JSONDecodeError:
            return json_str
    else:
        raise TypeError(f"Unsupported type for conversion from DynamoDB: {py_type}")


def convert_from_dynamo_attributes(
    fields: dict[str, FieldInfo], attributes: dict[str, Any]
) -> dict[str, Any]:
    """
    Convert the DynamoDB attributes to Python types.
    Each field is converted based on its type annotation.
    Python types are converted to DynamoDB types using convert_to_dynamo.
    This method is used when retrieving a model instance from DynamoDB.

    Args:
        attributes (dict[str, Any]): The DynamoDB attributes to be converted.

    Returns:
        dict[str, Any]: The converted attributes with Python types.
    """
    converted_attributes = {}
    for field_name, field_info in fields.items():
        field_value = attributes.get(field_name, {})
        field_type = field_info.annotation
        converted_attributes[field_name] = convert_from_dynamodb_format(
            field_value, field_type  # type: ignore
        )
    return converted_attributes


def convert_to_dynamo_attributes(
    fields: dict[str, FieldInfo], attributes: dict[str, Any]
) -> Dict[str, Any]:
    """
    Convert the instance to a DB-compatible dictionary.
    This method is used when saving a model instance to DynamoDB.

    Args:
        fields (dict[str, FieldInfo]): The fields of the model.
        attributes (dict[str, Any]): The DynamoDB attributes to be converted.

    Returns:
        dict[str, Any]: The DynamoDB-compatible dictionary.
    """
    dynamo_dict = {}

    for field_name, field_info in fields.items():
        field_type = field_info.annotation
        field_value = attributes[field_name]
        assert (
            field_info.json_schema_extra is not None
        ), f"Missing json_schema_extra for {field_name} in {fields}"
        use_map = field_info.json_schema_extra.get("use_map", False)
        if use_map:
            field_value = {k: convert_to_dynamo(v) for k, v in field_value.items()}
            dynamo_value = {"M": field_value}
        else:
            dynamo_value = convert_to_dynamodb_format(
                field_value,
                field_type,  # type: ignore
            )
        if dynamo_value is not None:
            dynamo_dict[field_name] = dynamo_value

    return dynamo_dict


class DynamoAdapter(DatabaseAdapter):
    client: Any
    table_name: str
    fields: Dict[str, Any]
    table_schema: Dict[str, Any]

    def __init__(
        self,
        client: Any,
        fields: Dict[str, Any],
        table_schema: Dict[str, Any],
    ):
        self.client = client
        self.fields = fields
        self.table_schema = table_schema
        self.table_name = table_schema["table_name"]

    def create_table(self) -> None:
        table_name = self.table_name
        key_schema = self.table_schema["key_schema"]
        attribute_definitions = self.table_schema["attribute_definitions"]
        global_secondary_indexes = self.table_schema.get("global_secondary_indexes", {})
        read_capacity_units = self.table_schema.get("read_capacity_units", 1)
        write_capacity_units = self.table_schema.get("write_capacity_units", 1)

        # Prepare key schema list
        key_schema_list = [
            {"AttributeName": name, "KeyType": key_type}
            for name, key_type in key_schema.items()
        ]

        # Prepare attribute definitions list
        attribute_definitions_list = [
            {"AttributeName": name, "AttributeType": type}
            for name, type in attribute_definitions.items()
        ]

        # Prepare global secondary indexes
        gsi_list = []
        if global_secondary_indexes:
            for index_name, schema in global_secondary_indexes.items():
                gsi_list.append(
                    {
                        "IndexName": index_name,
                        "KeySchema": [
                            {"AttributeName": name, "KeyType": key_type}
                            for name, key_type in schema.items()
                        ],
                        "Projection": {"ProjectionType": "ALL"},
                        "ProvisionedThroughput": {
                            "ReadCapacityUnits": read_capacity_units,
                            "WriteCapacityUnits": write_capacity_units,
                        },
                    }
                )

        # Table creation parameters
        table_params = {
            "TableName": table_name,
            "KeySchema": key_schema_list,
            "AttributeDefinitions": attribute_definitions_list,
            "ProvisionedThroughput": {
                "ReadCapacityUnits": read_capacity_units,
                "WriteCapacityUnits": write_capacity_units,
            },
        }

        if gsi_list:
            table_params["GlobalSecondaryIndexes"] = gsi_list

        dynamodb = Environment.get_dynamo_client()

        if table_name in dynamodb.list_tables()["TableNames"]:
            log.info(f"Table {table_name} already exists")
            return

        try:
            dynamodb.create_table(**table_params)
        except Exception as e:
            log.info(f"Table {table_name}: {e}")
            raise e
        else:
            log.info(f"Table {table_name} created")

    def drop_table(self) -> None:
        table_name = self.table_name
        dynamodb = Environment.get_dynamo_client()
        try:
            dynamodb.delete_table(TableName=table_name)
        except Exception as e:
            log.info(f"Table {table_name}: {e}")
            raise e
        else:
            log.info(f"Table {table_name} deleted")

    def get_primary_key(self) -> str:
        """
        Get the name of the hash key.
        """
        for field_name, field in self.fields.items():
            if field.json_schema_extra and field.json_schema_extra.get(
                "hash_key", False
            ):
                return field_name
        raise Exception(f"Hash key not found for {self.table_name}")

    def save(self, item: Dict[str, Any]) -> None:
        item = convert_to_dynamo_attributes(self.fields, item)
        try:
            self.client.put_item(TableName=self.table_name, Item=item)
        except Exception as e:
            log.error(f"Error saving {self.table_name}: {e} using {item}")
            raise e

    def get(self, key: Any) -> Dict[str, Any] | None:
        """
        Retrieve a model instance from DynamoDB using a key.
        The key can be a number, string.
        """

        try:
            dynamo_key = {
                self.get_primary_key(): {"S": str(key)},
            }
            response = self.client.get_item(TableName=self.table_name, Key=dynamo_key)
        except Exception as e:
            log.error(f"Error getting {self.table_name} with key {key}: {e}")
            raise e

        item = response.get("Item", None)
        if item is None:
            return None
        return convert_from_dynamo_attributes(self.fields, item)

    def delete(self, partition_value) -> None:
        self.client.delete_item(
            TableName=self.table_name,
            Key={
                self.get_primary_key(): {"S": partition_value},
            },
        )

    def query(
        self,
        condition: str,
        values: dict[str, Any],
        index: str | None = None,
        start_key: str | None = None,
        limit: int = 100,
        reverse: bool = False,
    ) -> tuple[list[dict[str, Any]], str]:
        """
        Query the DynamoDB table for the model to retrieve a list of items.
        This method is used for pagination and returns a tuple containing a list of items and the last evaluated key.
        It allows for filtering and sorting the results.

        Args:
            client: The DynamoDB client to use for the query.
            table_name: The name of the DynamoDB table to query.
            condition: The condition for the query based on the primary key.
            values: The values for the condition.
            index: The name of the index to use for the query.
            start_key: The exclusive start key for pagination.
            limit: The maximum number of items to retrieve.

        Returns:
            A tuple containing a list of items that match the query conditions and the last evaluated key.
        """
        # convert values to DynamoDB format
        values = {
            key: convert_to_dynamo(value) for key, value in values.items() if value
        }
        # Query parameters for DynamoDB query
        query_params = {
            "TableName": self.table_name,
            "KeyConditionExpression": condition,
            "ExpressionAttributeValues": values,
            "Limit": limit,
            "ScanIndexForward": not reverse,
        }
        if index:
            query_params["IndexName"] = index

        if start_key and start_key != "":
            query_params["ExclusiveStartKey"] = json.loads(start_key)

        # log.info("Querying DynamoDB: %s", query_params)

        try:
            response = self.client.query(**query_params)
        except Exception as e:
            log.error(f"Error querying {self.table_name}: {e}")
            raise e
        items = response.get("Items", [])

        # Extract the last evaluated key
        last_evaluated_key = (
            json.dumps(response.get("LastEvaluatedKey"))
            if response.get("LastEvaluatedKey")
            else ""
        )

        # Convert DynamoDB items to model objects
        items = [convert_from_dynamo_attributes(self.fields, item) for item in items]
        return items, last_evaluated_key
